"""
Keywords used across different PENELOPE main programs.
"""

# Standard library modules.
from operator import attrgetter

# Third party modules.

# Local modules.
from pypenelopetools.penelope.keyword import \
    TypeKeyword, KeywordSequence, KeywordGroup

# Globals and constants variables.

class TITLE(TypeKeyword):
    """Title of the job.
    
    The TITLE string is used to mark dump files. To prevent the
    improper use of wrong resuming files, change the title each
    time you modify basic parameters of your problem. The code
    will then be able to identify the inconsistency and to print
    an error message before stopping.
    """

    def __init__(self):
        super().__init__('TITLE', (str,))

    def _parse_line(self, line):
        name, values, comment = super()._parse_line(line)
        return name, (' '.join(values),), comment

    def set(self, title):
        """
        Sets value.
        
        Args:
            title (str): Title of the job (up to 65 characters).
        """
        super().set(title)

    def validate(self, title):
        if len(title) > 65:
            raise ValueError('Title is too long. Maximum 65 characters')
        return super().validate(title)

class SKPAR(TypeKeyword):
    """Type of primary particle KPARP (1=electrons, 2=photons or 3=positrons).
    
    If KPARP=0, the initial states of primary particles are
    set by subroutine SOURCE, to be provided by the user. An
    example of that subroutine, corresponding to a 60-Co source
    (two gamma rays in each nuclear deexcitation), is included
    in the PENMAIN package (file 'source.f').
    """

    def __init__(self):
        super().__init__('SKPAR', (int,),
                         comment="Primary particles: 1=electron, 2=photon, 3=positron")

    def set(self, kparp):
        """
        Sets value.
        
        Args:
            kparp (int): Type of primary particles.
                0=user defined, 1=electrons, 2=photons or 3=positrons
        """
        super().set(kparp)

    def validate(self, kparp):
        if kparp not in [0, 1, 2, 3]:
            raise ValueError('Invalid particle')
        return super().validate(kparp)

class SENERG(TypeKeyword):
    """For a monoenergetic source, initial energy SE0 of primary particles."""

    def __init__(self):
        super().__init__('SENERG', (float,),
                         comment="Initial energy (monoenergetic sources only)")

    def set(self, se0):
        """
        Sets value.
        
        Args:
            se0 (float): Initial energy of primary particles in eV
        """
        super().set(se0)

    def validate(self, se0):
        if se0 <= 0.0:
            raise ValueError('SE0 must be greater than 0')
        return super().validate(se0)

class SPECTR(KeywordSequence):
    """Define a source with continuous (stepwise constant) spectrum.
    
    For a source with continuous (stepwise constant) spectrum,
    each 'SPECTR' line gives the lower end-point of an energy
    bin of the source spectrum (Ei) and the associated relative
    probability (Pi), integrated over the bin. Up to NSEM=1000
    lines, in arbitrary order. The upper end of the spectrum is
    defined by entering a line with Ei equal to the upper energy
    end point and with a negative Pi value.
    """

    def __init__(self, maxlength=1000):
        keyword = TypeKeyword('SPECTR', (float, float),
                              comment='E bin: lower-end and total probability')
        super().__init__(keyword, maxlength)

    def add(self, ei, pi):
        """
        Adds a step in the spectrum.
        
        Args:
            ei (float): Lower end-point of an energy bin of the source spectrum in eV
            pi (float): Associated relative probability
        """
        return super().add(ei, pi)

class SGPOL(TypeKeyword):
    """Activates the simulation of polarisation effects in the scattering of photons.
    
    This line activates the simulation of polarisation effects
    in the scattering of photons (electrons and positrons are
    assumed to be unpolarised). SP1, SP2, SP3 are the Stokes
    parameters of primary photons, which define the degrees of
    linear polarisation at 45 deg azimuth, of circular
    polarisation, and of linear polarisation at zero azimuth,
    respectively. It is assumed that secondary photons are
    emitted with null polarisation (SP1=SP2=SP3=0).
    """

    def __init__(self):
        super().__init__('SGPOL', (float, float, float),
                         comment="Stokes parameters for polarized photons")

    def set(self, sp1, sp2, sp3):
        """
        Sets Stokes polarisation parameters.
        
        Args:
            sp1 (float): Degrees of linear polarisation at 45 deg azimuth
            sp2 (float): Degrees of circular polarisation
            sp3 (float): Degrees of linear polarisation at 0 deg azimuth
        """
        super().set(sp1, sp2, sp3)

class SPOSIT(TypeKeyword):
    """Coordinates of the source centre."""

    def __init__(self):
        super().__init__("SPOSIT", (float, float, float),
                         comment="Coordinates of the source")

    def set(self, sx0, sy0, sz0):
        """
        Sets coordinates.
        
        Args:
            sx0 (float): x-coordinate in cm.
            sy0 (float): y-coordinate in cm.
            sz0 (float): z-coordinate in cm.
        """
        super().set(sx0, sy0, sz0)

class SCONE(TypeKeyword):
    """Initial direction of primary particles is sampled uniformly within a conical beam.
    
    Conical source beam. Polar and azimuthal angles of the
    beam axis direction, THETA and PHI, and angular aperture,
    ALPHA, in deg.
    
    The case ALPHA=0 defines a monodirectional source, and ALPHA
    =180 deg corresponds to an isotropic source.
    """

    def __init__(self):
        super().__init__("SCONE", (float, float, float),
                         comment="Conical beam; angles in deg")

    def set(self, theta, phi, alpha):
        """
        Sets angles.
        
        Args:
            theta (float): Polar angle of the beam axis direction in deg.
            phi (float): Azimuthal angle of the beam axis direction in deg.
            alpha (float): Angular aperture in deg.
        """
        super().set(theta, phi, alpha)

class SRECTA(TypeKeyword):
    """Initial direction of primary particles is sampled uniformly within a rectangular beam.
    
    Rectangular source beam. Limiting polar and azimuthal angles
    of the source beam window, (THETAL,THETAU)x(PHIL,PHIU), in deg.
    
    The case THETAL=THETAU, PHIL=PHIU defines a monodirectional
    source. To define an isotropic source, set THETAL=0, THETAU=
    180, PHIL=0 and PHIU=360.
    """

    def __init__(self):
        super().__init__("SRECTA", (float, float, float, float),
                         comment="Rectangular beam; angles in deg")

    def set(self, thetal, thetau, phil, phiu):
        """
        Sets angles.
        
        Args:
            thetal (float): Lower limit polar angle in deg.
            thetau (float): Upper limit polar angle in deg.
            phil (float): Lower limit azimuthal angle in deg.
            phiu (float): Upper limit azimuthal angle in deg.
        """
        super().set(thetal, thetau, phil, phiu)

class MFNAME(TypeKeyword):
    """Name of a PENELOPE input material data file.
    
    This file must be generated in advance by running the program MATERIAL.
    """

    def __init__(self):
        super().__init__("MFNAME", (str,),
                         comment="Material file, up to 20 chars")

    def set(self, filename):
        """
        Sets filename.
        
        Args:
            filename (str): File name of material file (up to 20 characters).
        """
        super().set(filename)

class MSIMPA(TypeKeyword):
    """Set of simulation parameters for this material
    
    * absorption energies, EABS(1:3,M), 
    * elastic scattering parameters, C1(M) and C2(M), and 
    * cutoff energy losses for inelastic collisions and Bremsstrahlung emission, 
      WCC(M) and WCR(M).
    """

    def __init__(self):
        super().__init__("MSIMPA", (float, float, float, float, float, float, float),
                         comment="EABS(1:3),C1,C2,WCC,WCR")

    def set(self, eabs1, eabs2, eabs3, c1, c2, wcc, wcr):
        """
        Sets parameters.
        
        Args:
            eabs1 (float): Absorption energy of electrons in eV.
            eabs2 (float): Absorption energy of photons in eV.
            eabs3 (float): Absorption energy of positrons in eV.
            c1 (float): Elastic scattering coefficient.
            c2 (float): Elastic scattering coefficient.
            wcc (float): Cutoff energy losses for inelastic collisions in eV.
            wcr (float): Cutoff energy losses for Bremsstrahlung emission in eV.
        """
        super().set(eabs1, eabs2, eabs3, c1, c2, wcc, wcr)

class MaterialGroup(KeywordGroup):
    """Group to define both material file name and its simulation parameters."""

    def __init__(self):
        super().__init__()
        self.MFNAME = MFNAME()
        self.MSIMPA = MSIMPA()
        self.index = None

    def set(self, filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr, index=None):
        """
        Sets material file name and simulation parameters.
        
        Args:
            filename (str): File name of material file (up to 20 characters).
            eabs1 (float): Absorption energy of electrons in eV.
            eabs2 (float): Absorption energy of photons in eV.
            eabs3 (float): Absorption energy of positrons in eV.
            c1 (float): Elastic scattering coefficient.
            c2 (float): Elastic scattering coefficient.
            wcc (float): Cutoff energy losses for inelastic collisions in eV.
            wcr (float): Cutoff energy losses for Bremsstrahlung emission in eV.
            index (int, optional): Index of this material in the geometry
        """
        self.MFNAME.set(filename)
        self.MSIMPA.set(eabs1, eabs2, eabs3, c1, c2, wcc, wcr)
        self.index = index

    def get_keywords(self):
        return (self.MFNAME, self.MSIMPA)

class Materials(KeywordSequence):
    """Definition of materials."""

    def __init__(self, maxlength=10):
        keyword = MaterialGroup()
        super().__init__(keyword, maxlength)

    def add(self, index, filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr):
        """
        Adds a new material.
        
        Args:
            index (int): Index of this material in the geometry
            filename (str): File name of material file (up to 20 characters).
            eabs1 (float): Absorption energy of electrons in eV.
            eabs2 (float): Absorption energy of photons in eV.
            eabs3 (float): Absorption energy of positrons in eV.
            c1 (float): Elastic scattering coefficient.
            c2 (float): Elastic scattering coefficient.
            wcc (float): Cutoff energy losses for inelastic collisions in eV.
            wcr (float): Cutoff energy losses for Bremsstrahlung emission in eV.
        """
        return super().add(filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr, index)

    def _add_keyword(self, keyword):
        super()._add_keyword(keyword)
        if keyword.index is None:
            keyword.index = len(self._keywords)

    def get(self):
        values = []
        for keyword in sorted(self._keywords, key=attrgetter('index')):
            values.append(keyword.get())
        return (tuple(values),)

    def write(self, fileobj):
        for keyword in sorted(self._keywords, key=attrgetter('index')):
            keyword.write(fileobj)

class GEOMFN(TypeKeyword):
    """Name of geometry definition file.
    
    The bodies in the material structure are normally identified
    by the sequential labels assigned by PENGEOM. For complex
    geometries, however, it may be more practical to employ user
    labels, i.e., the four-character strings that identify the
    body in the geometry definition file. In PENMAIN (and only
    in the parts of the code that follow the definition of the
    geometry), a body can be specified by giving either its
    PENGEOM numerical label or its user label enclosed in a
    pair of apostrophes (e.g., 'BOD1'). However, bodies that
    result from the cloning of modules (as well as those defined
    in an INCLUDEd geometry file) do not have a user label and
    only the PENGEOM numerical label is acceptable.
    """

    def __init__(self):
        super().__init__("GEOMFN", (str,),
                         comment='Geometry file, up to 20 chars')

    def set(self, filename):
        """
        Sets filename.
        
        Args:
            filename (str): File name of material file (up to 20 characters).
        """
        super().set(filename)

class DSMAX(KeywordSequence):
    """Maximum step length of electrons and positrons in body. 
    
    .. note::
       This parameter is important only for thin bodies; it should be given a 
       value of the order of one tenth of the body thickness or less.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("DSMAX", (int, float),
                              comment="KB, maximum step length in body KB")
        super().__init__(keyword, maxlength)

    def add(self, kb, dsmax):
        """
        Sets maximum step length.
        
        Args:
            kb (int): Index of body.
            dsmax: Maximum step length in cm.
        """
        return super().add(kb, dsmax)

class EABSB(KeywordSequence):
    """Local absorption energies EABSB(KPAR,KB) of particles of type KPAR in body KB. 
    
    These values must be larger than EABS(KPAR,M), where M is the material of body KB. When the
    particle is moving within body KB, the absorption energy
    EABS(KPAR,M) is temporarily set equal to EABSB(KPAR,KB).
    Thus, the simulation of the particle history is discontinued
    when the energy becomes less than EABSB(KPAR,KB). This
    feature can be used, e.g., to reduce the simulation work in
    regions of lesser interest.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("EABSB", (int, float, float, float),
                              comment="KB, local absorption energies, EABSB(1:3)")
        super().__init__(keyword, maxlength)

    def add(self, kb, eabs1, eabs2, eabs3):
        """
        Sets local absorption energies.
        
        Args:
            kb (int): Index of body.
            eabs1 (float): Absorption energy of electrons in eV.
            eabs2 (float): Absorption energy of photons in eV.
            eabs3 (float): Absorption energy of positrons in eV.
        """
        return super().add(kb, eabs1, eabs2, eabs3)

class IFORCE(KeywordSequence):
    """Forcing of interactions. 
    
    FORCER is the forcing factor, which must
    be larger than unity. WLOW and WHIG are the lower and upper
    limits of the weight window where interaction forcing is
    applied. When several interaction mechanisms are forced in
    the same body, the effective weight window is set equal to
    the intersection of the windows for these mechanisms.
    
    If the mean free path for real interactions of type ICOL is
    MFP, the program will simulate interactions of this type
    (real or forced) with an effective mean free path equal to
    MFP/FORCER.
    
    .. hint::
       A negative input value of FORCER, -FN, is assumed to mean that a particle 
       with energy E=EPMAX should interact, on average, +FN times in the course 
       of its slowing down to rest, for electrons and positrons, or along a mean
       free path, for photons. This is very useful, e.g., to generate x-ray 
       spectra from bulk samples.
    """

    def __init__(self, maxlength=120000):
        keyword = TypeKeyword("IFORCE", (int, int, int, float, float, float),
                              comment="KB,KPAR,ICOL,FORCER,WLOW,WHIG")
        super().__init__(keyword, maxlength)

    def add(self, kb, kpar, icol, forcer, wlow, whig):
        """
        Adds forcing for an interaction.
        
        Args:
            kb (int): Index of body.
            kparp (int): Type of primary particles.
                0=user defined, 1=electrons, 2=photons or 3=positrons
        """
        return super().add(kb, kpar, icol, forcer, wlow, whig)

class IBRSPL(KeywordSequence):
    """Bremsstrahlung splitting for electrons and positrons.
    
    .. note:: 
       Note that bremsstrahlung splitting is applied in combination
       with interaction forcing and, consequently, it is activated
       only in those bodies where interaction forcing is active.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("IBRSPL", (int, float),
                              comment="KB,splitting factor")
        super().__init__(keyword, maxlength)

    def add(self, kb, ibrspl):
        """
        Add Bremsstrahlung splitting.
        
        Args:
            kb (int): Index of body.
            ibrspl (int): Splitting factor.
        """
        return super().add(kb, ibrspl)

class IXRSPL(KeywordSequence):
    """Splitting of characteristic x rays emitted. 
    
    Each unsplit x ray with ILB(2)=2 (i.e., of the second generation) when 
    extracted from the secondary stack is split into IXRSPL quanta. 
    The new, lighter, quanta are assigned random directions distributed 
    isotropically.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("IXRSPL", (int, float),
                              comment="KB,splitting factor")
        super().__init__(keyword, maxlength)

    def add(self, kb, ixrspl):
        """
        Add characteristic x rays splitting.
        
        Args:
            kb (int): Index of body.
            ixrspl (int): Splitting factor.
        """
        return super().add(kb, ixrspl)

class NBE(TypeKeyword):
    """Definition of energy distributions of emerging particles."""

    def __init__(self):
        super().__init__("NBE", (float, float, int),
                         comment="Energy window and no. of bins")

    def set(self, el, eu, nbe):
        """
        Sets energy distributions.
        
        Args:
            el (float): Lower limit in eV.
            eu (float): Upper limit in eV.
            nbe (int): Number of bins in the output energy distribution.
                Should be less than 1000. 
                If NBE is positive, energy bins have uniform width, 
                DE=(EU-EL)/NBE. 
                When NBE is negative, the bin width increases geometrically 
                with the energy, i.e., the energy bins have uniform width on a 
                logarithmic scale.
        """
        super().set(el, eu, nbe)

class NBANGL(TypeKeyword):
    """Definition of angular distributions of emerging particles.
    
    .. note::
       In the output files, the terms 'upbound' and 'downbound' are used to 
       denote particles that leave the material system moving upwards (W>0) and 
       downwards (W<0), respectively.
    """

    def __init__(self):
        super().__init__("NBANGL", (int, int),
                         comment="No. of bins for the angles THETA and PHI")

    def set(self, nbth, nbph):
        """
        Sets angular distributions.
        
        Args:
            nbth (int): Numbers of bins for the polar angle THETA. 
                Should be less than 3600.
                If NBTH is positive, angular bins have uniform width,
                DTH=180./NBTHE. 
                When NBTH is negative, the bin width increases geometrically 
                with THETA, i.e., the bins have uniform width on a logarithmic 
                scale.
            nbph (int): Number of bins for the azimuthal angle PHI
                Should be less than 180.
        """
        super().set(nbth, nbph)

#TODO: Fix ENDETC, EDSPC. It should be a KeywordSequence
class ENDETC(TypeKeyword):
    """Definition of an energy-deposition detector."""

    def __init__(self):
        super().__init__('ENDETC', (float, float, int),
                         comment='Energy window and no. of bins')

    def set(self, el, eu, nbe):
        """
        Sets energy limits.
        
        Args:
            el (float): Lower limit in eV.
            eu (float): Upper limit in eV.
            nbe (int): Number of bins in the output energy distribution.
                Should be less than 1000. 
                If NBE is positive, energy bins have uniform width, 
                DE=(EU-EL)/NBE. 
                When NBE is negative, the bin width increases geometrically 
                with the energy, i.e., the energy bins have uniform width on a 
                logarithmic scale.
        """
        super().set(el, eu, nbe)

class EDSPC(TypeKeyword):
    """Name of the output spectrum file."""

    def __init__(self):
        super().__init__('EDSPC', (str,),
                         comment='Output spectrum file name, 20 chars')

    def set(self, filename):
        """
        Sets filename.
        
        Args:
            filename (str): File name of output spectrum file (up to 20 characters).
        """
        super().set(filename)

class GRIDX(TypeKeyword):
    """Definition of x-coordinates of the vertices of the dose box."""

    def __init__(self):
        super().__init__('GRIDX', (float, float, int),
                         comment='X coords of the box vertices, no. of bins')

    def set(self, xl, xu, ndbx):
        """
        Sets dimensions.
        
        Args:
            xl (float): Lower limit of the dose box along the x-axis in cm.
            xu (float): Upper limit of the dose box along the x-axis in cm.
            ndbx (int): Number of bins (i.e. voxels) along the x-axis.
        """
        super().set(xl, xu, ndbx)

class GRIDY(TypeKeyword):
    """Definition of y-coordinates of the vertices of the dose box."""

    def __init__(self):
        super().__init__('GRIDY', (float, float, int),
                         comment='Y coords of the box vertices, no. of bins')

    def set(self, yl, yu, ndby):
        """
        Sets dimensions.
        
        Args:
            yl (float): Lower limit of the dose box along the y-axis in cm.
            yu (float): Upper limit of the dose box along the y-axis in cm.
            ndby (int): Number of bins (i.e. voxels) along the y-axis.
        """
        super().set(yl, yu, ndby)

class GRIDZ(TypeKeyword):
    """Definition of z-coordinates of the vertices of the dose box."""

    def __init__(self):
        super().__init__('GRIDZ', (float, float, int),
                         comment='Z coords of the box vertices, no. of bins')

    def set(self, zl, zu, ndbz):
        """
        Sets dimensions.
        
        Args:
            zl (float): Lower limit of the dose box along the z-axis in cm.
            zu (float): Upper limit of the dose box along the z-axis in cm.
            ndbz (int): Number of bins (i.e. voxels) along the z-axis.
        """
        super().set(zl, zu, ndbz)

class RESUME(TypeKeyword):
    """Name of the resume file.
    
    The program will read the dump file named ``dump1.dmp`` and resume the 
    simulation from the point where it was left. 
    
    .. warning::
       Use this option very, **VERY** carefully.
       Make sure that the input data file is fully consistent with the one used 
       to generate the dump file.
    """

    def __init__(self):
        super().__init__('RESUME', (str,),
                         comment='Resume from this dump file, 20 chars')

    def set(self, filename):
        """
        Sets filename.
        
        Args:
            filename (str): File name of resume file (up to 20 characters).
        """
        super().set(filename)

class DUMPTO(TypeKeyword):
    """Name of dump file.
    
    Generate a dump file named 'dump2.dmp' after completing the simulation run. 
    This allows the simulation to be resumed later on to improve statistics.
    
    .. note:: 
       If the file 'dump2.dmp' already exists, it is overwritten.
    """

    def __init__(self):
        super().__init__('DUMPTO', (str,),
                         comment='Generate this dump file, 20 chars')

    def set(self, filename):
        """
        Sets filename.
        
        Args:
            filename (str): File name of dump file (up to 20 characters).
        """
        super().set(filename)

class DUMPP(TypeKeyword):
    """Dump interval.
    
    When the DUMPTO option is activated, simulation results are written in the 
    output files every DUMPP seconds. 
    This option is useful to check the progress of long simulations. 
    It also allows the program to be run with a long execution time and to be 
    stopped when the required statistical uncertainty has been reached.
    """

    def __init__(self):
        super().__init__('DUMPP', (float,), comment='Dumping period, in sec')

    def set(self, dumpp):
        """
        Sets interval.
        
        Args:
            dumpp (int): Dump interval in seconds.
        """
        super().set(dumpp)

class RSEED(TypeKeyword):
    """Seeds of the random-number generator."""

    def __init__(self):
        super().__init__('RSEED', (int, int),
                         comment='Seeds of the random-number generator')

    def set(self, iseed1, iseed2):
        """
        Sets seeds.
        
        Args:
            iseed1 (int): First seed.
                When ISEED1 is equal to a negative integer, -N, the seeds are 
                set by calling subroutine RAND0(N) with the input argument 
                equal to N.
                This ensures that sequences of random numbers used in different
                runs of the program (with different values of N) are truly
                independent.
            iseed2 (int): Second seed.
        """
        super().set(iseed1, iseed2)

class NSIMSH(TypeKeyword):
    """Desired number of simulated showers."""

    def __init__(self):
        super().__init__('NSIMSH', (float,),
                         comment='Desired number of simulated showers')

    def set(self, dshn):
        """
        Sets showers.
        
        Args:
            dshn (int): Number of simulated showers.
        """
        super().set(dshn)

class TIME(TypeKeyword):
    """Allotted simulation time."""

    def __init__(self):
        super().__init__('TIME', (float,),
                         comment='Allotted simulation time, in sec')

    def set(self, timea):
        """
        Sets simulation time.
        
        Args:
            timea (int): Allotted simulation time in seconds.
        """
        super().set(timea)




"""
Keywords used specifically for PENEPMA.
"""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.keyword import \
    TypeKeyword, KeywordGroup, KeywordSequence

# Globals and constants variables.

class SRADI(TypeKeyword):
    """Initial position of the particle is sampled randomly within a circle.
    
    The circle is centered at (SX0,SY0,SZ0) and perpendicular to the beam 
    axis direction.
    """

    def __init__(self):
        super().__init__('SRADI', (float,),
                         comment='Radius of the beam, in cm')

    def set(self, sradius):
        """
        Sets circle.
        
        Args:
            sradius (float): Radius of the circle in cm.
        """
        super().set(sradius)

class SDIREC(TypeKeyword):
    """Polar and azimuthal angles of the electron beam axis direction."""

    def __init__(self):
        super().__init__('SDIREC', (float, float),
                         comment='Direction angles of the beam axis, in deg')

    def set(self, stheta, sphi):
        """
        Sets angles.
        
        Args:
            stheta (float): Polar angle in deg.
            sphi (float): Azimuthal angle in deg.
        """
        super().set(stheta, sphi)

class SAPERT(TypeKeyword):
    """Angular aperture of the electron beam."""

    def __init__(self):
        super().__init__('SAPERT', (float,), comment='Beam aperture, in deg')

    def set(self, salpha):
        """
        Sets angle.
        
        Args:
            salpha (float): Angular aperture in deg.
        """
        super().set(salpha)

class PhotonDetectorGroup(KeywordGroup):
    """Definition of photon detector.
    
    Each detector collects photons that leave the sample with directions within 
    a *rectangle( on the unit sphere, limited by the *parallels* THETA1 and 
    THETA2 and the *meridians* PHI1 and PHI2. 
    The output spectrum is the energy distribution of photons that emerge within 
    the acceptance solid angle of the detector with energies in the interval 
    from EDEL to EDEU, recorded using NCHE channels. 
    Notice that the spectrum is given in absolute units (per incident electron, 
    per eV and per unit solid angle).
    """

    def __init__(self):
        super().__init__()
        self.PDANGL = TypeKeyword('PDANGL', (float, float, float, float, int),
                                  comment='Angular window, in deg, IPSF')
        self.PDENER = TypeKeyword('PDENER', (float, float, int),
                                  comment='Energy window, no. of channels')
        self.XRORIG = TypeKeyword('XRORIG', (str,),
                                  comment='Map of emission sites of detected x rays')

    def set(self, theta1, theta2, phi1, phi2, ipsf, edel, edeu, nche, emission_filename=None):
        """
        Sets parameters of detector.
        
        .. note::
           ``phi1`` and ``phi2`` must be both either in the interval (0,360) or 
           in the interval (-180,180).
        
        Args:
            theta1 (float): Lower limit polar angle in deg.
            theta2 (float): Upper limit polar angle in deg.
            phi1 (float): Lower limit azimuthal angle in deg.
            phi2 (float): Upper limit azimuthal angle in deg.
            ipsf (int): Flag to activate the creation of a phase-space file
                (psf), which contains the state variables and weights of 
                particles that enter the detector. 
                Use this option with care, because psf's may grow very fast.
                
                * ``ipsf=0``: the psf is not created.
                * ``ipsf=1``: a psf is created.
                * ``ipsf>1``: a psf is created, but contains only state
                  variables of detected photons that have ILB(4)=IPSF 
                  (used for studying angular distributions of x rays).
                
                Generating the psf is useful for tuning interaction forcing,
                which requires knowing the weights of detected particles.
            edel (float): Lower limits of the energy window covered by the 
                detector in eV.
            edeu (float): Upper limits of the energy window covered by the 
                detector in eV.
            nche (int): Number of energy channels in the output spectrum.
                Should be less than 1000.
            emission_filename (str, optional): File name of the generation file.
                Specifying a file name activates the generation of a file with 
                the position coordinates of the emission sites of the photons 
                that reach the detector. 
                Notice that the file may grow very fast, so use this option 
                only in short runs. 
                The output file is overwritten when a simulation is resumed.
        """
        self.PDANGL.set(theta1, theta2, phi1, phi2, ipsf)
        self.PDENER.set(edel, edeu, nche)
        self.XRORIG.set(emission_filename)

    def get_keywords(self):
        return (self.PDANGL, self.PDENER, self.XRORIG)

class PhotonDetectors(KeywordSequence):
    """Definition of the photon detectors.
    
    Up to 25 different detectors can be defined.
    """

    def __init__(self, maxlength=25):
        keyword = PhotonDetectorGroup()
        super().__init__(keyword, maxlength)

    def add(self, theta1, theta2, phi1, phi2, ipsf, edel, edeu, nche, emission_filename=None):
        """
        Add a new photon detector.
        
        .. note::
           ``phi1`` and ``phi2`` must be both either in the interval (0,360) or 
           in the interval (-180,180).
           
        .. important::
           The theta angles are defined as angles from the positive z-axis.
           This is different than the take-off angle usually used in 
           microanalysis.
           For a take-off angle of 30deg, theta would be 60deg.
        
        Args:
            theta1 (float): Lower limit polar angle in deg.
            theta2 (float): Upper limit polar angle in deg.
            phi1 (float): Lower limit azimuthal angle in deg.
            phi2 (float): Upper limit azimuthal angle in deg.
            ipsf (int): Flag to activate the creation of a phase-space file
                (psf), which contains the state variables and weights of 
                particles that enter the detector. 
                Use this option with care, because psf's may grow very fast.
                
                * ``ipsf=0``: the psf is not created.
                * ``ipsf=1``: a psf is created.
                * ``ipsf>1``: a psf is created, but contains only state
                  variables of detected photons that have ILB(4)=IPSF 
                  (used for studying angular distributions of x rays).
                
                Generating the psf is useful for tuning interaction forcing,
                which requires knowing the weights of detected particles.
            edel (float): Lower limits of the energy window covered by the 
                detector in eV.
            edeu (float): Upper limits of the energy window covered by the 
                detector in eV.
            nche (int): Number of energy channels in the output spectrum.
                Should be less than 1000.
            emission_filename (str, optional): File name of the generation file.
                Specifying a file name activates the generation of a file with 
                the position coordinates of the emission sites of the photons 
                that reach the detector. 
                Notice that the file may grow very fast, so use this option 
                only in short runs. 
                The output file is overwritten when a simulation is resumed.
        """
        super().add(theta1, theta2, phi1, phi2, ipsf, edel, edeu, nche, emission_filename)

class XRAYE(KeywordSequence):
    """Space distribution of emission sites of x rays within an energy interval."""

    def __init__(self, maxlength=10):
        keyword = TypeKeyword('XRAYE', (float, float),
                              comment='Energy interval where x rays are mapped')
        super().__init__(keyword, maxlength)

    def add(self, emin, emax):
        """
        Add an energy interval.
        
        Args:
            emin (float): Lower limit in eV.
            emax (float): Upper limit in eV.
        """
        super().add(emin, emax)

class XRLINE(KeywordSequence):
    """Space distribution of emission sites of x rays."""

    def __init__(self, maxlength=10):
        keyword = TypeKeyword('XRLINE', (int,),
                              comment='X-ray line, IZ*1e6+S1*1e4+S2*1e2')
        super().__init__(keyword, maxlength)

    def add(self, izs1s200):
        """
        Adds a characteristic x ray.
        
        .. hint::
           Use :func:`convert_xrayline_to_izs1s200 <pypenelopetools.penepma.utils.convert_xrayline_to_izs1s200>`
           to convert :class:`XrayLine` to ILB(4) notation.
        
        Args:
            izs1s200 (int): x ray identification.
                ILB(4) notation, note the final double zero (IZ*1e6+S1*1e4+S2*1e2).
        """
        super().add(izs1s200)

class REFLIN(TypeKeyword):
    """Termination of simulation based on relative statistical uncertainty of the intensity of line."""

    def __init__(self):
        super().__init__('REFLIN', (int, int, float),
                         comment='IZ*1e6+S1*1e4+S2*1e2,detector,tolerance')

    def set(self, izs1s200, idet, tol):
        """
        Sets termination.
        
        .. hint::
           Use :func:`convert_xrayline_to_izs1s200 <pypenelopetools.penepma.utils.convert_xrayline_to_izs1s200>`
           to convert :class:`XrayLine` to ILB(4) notation.
        
        Args:
            izs1s200: x ray identification.
                ILB(4) notation, note the final double zero (IZ*1e6+S1*1e4+S2*1e2).
            idet (int): Index of detector.
            tol (float): Relative statistical uncertainty (3*sigma) of the intensity of line
        """
        super().set(izs1s200, idet, tol)






""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.input import _PenelopeInputBase
import pypenelopetools.penelope.keywords as penelope_keywords
import pypenelopetools.penelope.separators as penelope_separators
import pypenelopetools.pencyl.keywords as pencyl_keywords
import pypenelopetools.pencyl.separators as pencyl_separators

# Globals and constants variables.

class PencylInput(_PenelopeInputBase):

    def __init__(self):
        self.TITLE = penelope_keywords.TITLE()

        # Geometry definition list
        self.geometry_definitions = pencyl_keywords.GeometryDefinitions()

        # Source definition
        self.SKPAR = penelope_keywords.SKPAR()
        self.SENERG = penelope_keywords.SENERG()
        self.SPECTR = penelope_keywords.SPECTR()
        self.SGPOL = penelope_keywords.SGPOL()
        self.SEXTND = pencyl_keywords.SEXTND()
        self.STHICK = pencyl_keywords.STHICK()
        self.SRADII = pencyl_keywords.SRADII()
        self.SPOSIT = penelope_keywords.SPOSIT()
        self.SCONE = penelope_keywords.SCONE()
        self.SRECTA = penelope_keywords.SRECTA()

        # Material data and simulation parameters
        self.materials = penelope_keywords.Materials()

        # Geometry and local simulation parameters
        self.DSMAX = pencyl_keywords.DSMAX()
        self.EABSB = pencyl_keywords.EABSB()

        # Interaction forcing
        self.IFORCE = pencyl_keywords.IFORCE()

        # Bremsstrahlung splitting
        self.IBRSPL = pencyl_keywords.IBRSPL()

        # X-ray splitting
        self.IXRSPL = pencyl_keywords.IXRSPL()

        # Woodcock's delta-scattering method for photons.
        self.IWOODC = pencyl_keywords.IWOODC()

        # Emerging particles
        self.NBE = penelope_keywords.NBE()
        self.NBANGL = penelope_keywords.NBANGL()
        self.NBZ = pencyl_keywords.NBZ()
        self.NBR = pencyl_keywords.NBR()
        self.NBTL = pencyl_keywords.NBTL()

        # Particle positions at the lower and upper planes
        self.EMERGP = pencyl_keywords.EMERGP()

        # Energy-deposition detectors
        self.energy_deposition_detectors = pencyl_keywords.EnergyDepositionDetectors()

        # Dose and charge distributions
        self.DOSE2D = pencyl_keywords.DOSE2D()

        # Job properties
        self.RESUME = penelope_keywords.RESUME()
        self.DUMPTO = penelope_keywords.DUMPTO()
        self.DUMPP = penelope_keywords.DUMPP()
        self.RSEED = penelope_keywords.RSEED()
        self.NSIMSH = penelope_keywords.NSIMSH()
        self.TIME = penelope_keywords.TIME()

    def get_keywords(self):
        return [self.TITLE,

                penelope_separators.DOT,
                pencyl_separators.GEOMETRY_LIST_START,
                self.geometry_definitions,
                pencyl_separators.GEOMETRY_LIST_END,

                penelope_separators.DOT,
                penelope_separators.SOURCE_DEFINITION,
                self.SKPAR, self.SENERG, self.SPECTR, self.SGPOL,
                self.SEXTND, self.STHICK, self.SRADII,
                self.SPOSIT, self.SCONE, self.SRECTA,

                penelope_separators.DOT,
                penelope_separators.MATERIAL,
                self.materials,

                penelope_separators.DOT,
                pencyl_separators.DSMAX_EABSB,
                self.DSMAX, self.EABSB,

                penelope_separators.DOT,
                penelope_separators.INTERACTION_FORCING,
                self.IFORCE,

                penelope_separators.DOT,
                penelope_separators.BREMSSTRAHLUNG_SPLITTING,
                self.IBRSPL,

                penelope_separators.DOT,
                penelope_separators.XRAY_SPLITTING,
                self.IXRSPL,

                penelope_separators.DOT,
                pencyl_separators.WOODCOCK,
                self.IWOODC,

                penelope_separators.DOT,
                pencyl_separators.COUNTER_ARRAY,
                self.NBE, self.NBANGL, self.NBZ, self.NBR, self.NBTL,

                penelope_separators.DOT,
                pencyl_separators.PARTICLE_POSITIONS,
                self.EMERGP,

                penelope_separators.DOT,
                penelope_separators.ENERGY_DEPOSITON_DETECTORS,
                self.energy_deposition_detectors,

                penelope_separators.DOT,
                pencyl_separators.DOSE_CHARGE_DISTRIBUTION,
                self.DOSE2D,

                penelope_separators.DOT,
                penelope_separators.JOB_PROPERTIES,
                self.RESUME, self.DUMPTO, self.DUMPP,

                penelope_separators.DOT,
                self.RSEED, self.NSIMSH, self.TIME,

                penelope_separators.END]



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import io
import os

# Third party modules.
import pyxray

# Local modules.
from pypenelopetools.penepma.input import PenepmaInput
from pypenelopetools.material import Material
from pypenelopetools.pengeom.surface import xplane, zplane, cylinder
from pypenelopetools.pengeom.module import Module, SidePointer
from pypenelopetools.pengeom.geometry import Geometry
from pypenelopetools.penepma.utils import convert_xrayline_to_izs1s200

# Globals and constants variables.

MATERIAL_CU = Material('Cu', {29: 1.0}, 8.9)
MATERIAL_FE = Material('Fe', {26: 1.0}, 7.874)
XRAYLINE_CU_KA2 = pyxray.xray_line(29, 'Ka2')
XRAYLINE_FE_KA2 = pyxray.xray_line(26, 'Ka2')

def create_epma1():
    # Create geometry
    module = Module(MATERIAL_CU, 'Sample')
    module.add_surface(zplane(0.0), SidePointer.NEGATIVE)
    module.add_surface(zplane(-0.1), SidePointer.POSITIVE)
    module.add_surface(cylinder(1.0), SidePointer.NEGATIVE)

    geometry = Geometry('Cylindrical homogeneous foil')
    geometry.add_module(module)

    index_lookup = geometry.indexify()

    # Create input
    input = PenepmaInput()

    input.TITLE.set('A Cu slab')
    input.SENERG.set(15e3)
    input.SPOSIT.set(0.0, 0.0, 1.0)
    input.SDIREC.set(180, 0.0)
    input.SAPERT.set(0.0)

    input.materials.add(index_lookup[MATERIAL_CU], MATERIAL_CU.filename, 1e3, 1e3, 1e3, 0.2, 0.2, 1e3, 1e3)

    input.GEOMFN.set('epma1.geo')
    input.DSMAX.add(index_lookup[module], 1e-4)

    input.IFORCE.add(index_lookup[module], 1, 4, -5, 0.9, 1.0)
    input.IFORCE.add(index_lookup[module], 1, 5, -250, 0.9, 1.0)
    input.IFORCE.add(index_lookup[module], 2, 2, 10, 1e-3, 1.0)
    input.IFORCE.add(index_lookup[module], 2, 3, 10, 1e-3, 1.0)
    input.IBRSPL.add(index_lookup[module], 2)
    input.IXRSPL.add(index_lookup[module], 2)

    input.NBE.set(0, 0, 300)
    input.NBANGL.set(45, 30)

    input.photon_detectors.add(0, 90, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(5, 15, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(15, 25, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(25, 35, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(35, 45, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(45, 55, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(55, 65, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(65, 75, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(75, 85, 0, 360, 0, 0.0, 0.0, 1000)

    input.GRIDX.set(-4e-5, 4e-5, 60)
    input.GRIDY.set(-4e-5, 4e-5, 60)
    input.GRIDZ.set(-6e-5, 0.0, 60)
    input.XRLINE.add(convert_xrayline_to_izs1s200(XRAYLINE_CU_KA2))

    input.RESUME.set('dump1.dat')
    input.DUMPTO.set('dump1.dat')
    input.DUMPP.set(60)

    input.RSEED.set(-10, 1)
    input.REFLIN.set(convert_xrayline_to_izs1s200(XRAYLINE_CU_KA2), 1, 1.25e-3)
    input.NSIMSH.set(2e9)
    input.TIME.set(2e9)

    return input

def create_epma2():
    # Create geometry
    surface_top = zplane(0.0)
    surface_bottom = zplane(-0.1)
    surface_cylinder = cylinder(1.0)
    surface_divider = xplane(0.0)

    module_right = Module(MATERIAL_CU, 'Right half of the sample')
    module_right.add_surface(surface_top, SidePointer.NEGATIVE)
    module_right.add_surface(surface_bottom, SidePointer.POSITIVE)
    module_right.add_surface(surface_cylinder, SidePointer.NEGATIVE)
    module_right.add_surface(surface_divider, SidePointer.POSITIVE)

    module_left = Module(MATERIAL_FE, 'Left half of the sample')
    module_left.add_surface(surface_top, SidePointer.NEGATIVE)
    module_left.add_surface(surface_bottom, SidePointer.POSITIVE)
    module_left.add_surface(surface_cylinder, SidePointer.NEGATIVE)
    module_left.add_module(module_right)

    geometry = Geometry('Cylindrical homogeneous foil')
    geometry.add_module(module_right)
    geometry.add_module(module_left)

    index_lookup = geometry.indexify()
    index_lookup[MATERIAL_CU] = 1
    index_lookup[MATERIAL_FE] = 2

    # Create input
    input = PenepmaInput()

    input.TITLE.set('A CU-Fe couple')
    input.SENERG.set(15e3)
    input.SPOSIT.set(2e-5, 0.0, 1.0)
    input.SDIREC.set(180, 0.0)
    input.SAPERT.set(0.0)

    input.materials.add(index_lookup[MATERIAL_FE], MATERIAL_FE.filename, 1e3, 1e3, 1e3, 0.2, 0.2, 1e3, 1e3) # Note inversion
    input.materials.add(index_lookup[MATERIAL_CU], MATERIAL_CU.filename, 1e3, 1e3, 1e3, 0.2, 0.2, 1e3, 1e3)

    input.GEOMFN.set('epma2.geo')
    input.DSMAX.add(index_lookup[module_right], 1e-4)
    input.DSMAX.add(index_lookup[module_left], 1e-4)

    input.IFORCE.add(index_lookup[module_right], 1, 4, -5, 0.9, 1.0)
    input.IFORCE.add(index_lookup[module_right], 1, 5, -250, 0.9, 1.0)
    input.IFORCE.add(index_lookup[module_right], 2, 2, 10, 1e-3, 1.0)
    input.IFORCE.add(index_lookup[module_right], 2, 3, 10, 1e-3, 1.0)
    input.IFORCE.add(index_lookup[module_left], 1, 4, -5, 0.9, 1.0)
    input.IFORCE.add(index_lookup[module_left], 1, 5, -7, 0.9, 1.0)
    input.IFORCE.add(index_lookup[module_left], 2, 2, 10, 1e-3, 1.0)
    input.IFORCE.add(index_lookup[module_left], 2, 3, 10, 1e-3, 1.0)

    input.IBRSPL.add(index_lookup[module_right], 2)
    input.IBRSPL.add(index_lookup[module_left], 2)

    input.IXRSPL.add(index_lookup[module_right], 2)
    input.IXRSPL.add(index_lookup[module_left], 2)

    input.NBE.set(0, 0, 300)
    input.NBANGL.set(45, 30)

    input.photon_detectors.add(0, 90, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(5, 15, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(15, 25, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(25, 35, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(35, 45, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(45, 55, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(55, 65, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(65, 75, 0, 360, 0, 0.0, 0.0, 1000)
    input.photon_detectors.add(75, 85, 0, 360, 0, 0.0, 0.0, 1000)

    input.GRIDX.set(-1e-5, 5e-5, 60)
    input.GRIDY.set(-3e-5, 3e-5, 60)
    input.GRIDZ.set(-6e-5, 0.0, 60)
    input.XRLINE.add(convert_xrayline_to_izs1s200(XRAYLINE_FE_KA2))
    input.XRLINE.add(convert_xrayline_to_izs1s200(XRAYLINE_CU_KA2))

    input.RESUME.set('dump2.dat')
    input.DUMPTO.set('dump2.dat')
    input.DUMPP.set(60)

    input.RSEED.set(-10, 1)
    input.REFLIN.set(convert_xrayline_to_izs1s200(XRAYLINE_FE_KA2), 1, 1.5e-3)
    input.NSIMSH.set(2e9)
    input.TIME.set(2e9)

    return input

class TestPenmainInput(unittest.TestCase):

    def setUp(self):
        super().setUp()

        self.testdatadir = os.path.join(os.path.dirname(__file__),
                                        '..', 'testdata', 'penepma')

    def _write_read_input(self, input):
        fileobj = io.StringIO()

        try:
            input.write(fileobj)

            fileobj.seek(0)
            outinput = PenepmaInput()
            outinput.read(fileobj)
        finally:
            fileobj.close()

        return outinput

    def _test_epma1(self, input):
        se0, = input.SENERG.get()
        self.assertAlmostEqual(15e3, se0, 5)

        sx0, sy0, sz0 = input.SPOSIT.get()
        self.assertAlmostEqual(0.0, sx0, 5)
        self.assertAlmostEqual(0.0, sy0, 5)
        self.assertAlmostEqual(1.0, sz0, 5)

        theta, phi = input.SDIREC.get()
        self.assertAlmostEqual(180.0, theta, 5)
        self.assertAlmostEqual(0.0, phi, 5)

        alpha, = input.SAPERT.get()
        self.assertAlmostEqual(0.0, alpha, 5)

        materials, = input.materials.get()
        self.assertEqual(1, len(materials))

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[0]
        self.assertEqual('Cu.mat', filename)
        self.assertAlmostEqual(1e3, eabs1, 5)
        self.assertAlmostEqual(1e3, eabs2, 5)
        self.assertAlmostEqual(1e3, eabs3, 5)
        self.assertAlmostEqual(0.2, c1, 5)
        self.assertAlmostEqual(0.2, c2, 5)
        self.assertAlmostEqual(1e3, wcc, 5)
        self.assertAlmostEqual(1e3, wcr, 5)

        self.assertEqual('epma1.geo', input.GEOMFN.get()[0])

        dsmaxs, = input.DSMAX.get()
        self.assertEqual(1, len(dsmaxs))

        kb, dsmax = dsmaxs[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(1e-4, dsmax, 5)

        iforces, = input.IFORCE.get()
        self.assertEqual(4, len(iforces))

        kb, kpar, icol, forcer, wlow, whig = iforces[0]
        self.assertEqual(1, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(4, icol)
        self.assertAlmostEqual(-5, forcer, 5)
        self.assertAlmostEqual(0.9, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[1]
        self.assertEqual(1, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(5, icol)
        self.assertAlmostEqual(-250, forcer, 5)
        self.assertAlmostEqual(0.9, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[2]
        self.assertEqual(1, kb)
        self.assertEqual(2, kpar)
        self.assertEqual(2, icol)
        self.assertAlmostEqual(10, forcer, 5)
        self.assertAlmostEqual(1e-3, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[3]
        self.assertEqual(1, kb)
        self.assertEqual(2, kpar)
        self.assertEqual(3, icol)
        self.assertAlmostEqual(10, forcer, 5)
        self.assertAlmostEqual(1e-3, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        ibrspls, = input.IBRSPL.get()
        self.assertEqual(1, len(ibrspls))

        kb, factor = ibrspls[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        ixrspls, = input.IXRSPL.get()
        self.assertEqual(1, len(ixrspls))

        kb, factor = ixrspls[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        el, eu, nbe = input.NBE.get()
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(300, nbe)

        nbth, nbph = input.NBANGL.get()
        self.assertEqual(45, nbth)
        self.assertEqual(30, nbph)

        photon_detectors, = input.photon_detectors.get()
        self.assertEqual(9, len(photon_detectors))

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[0]
        self.assertAlmostEqual(0.0, theta1, 5)
        self.assertAlmostEqual(90.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[1]
        self.assertAlmostEqual(5.0, theta1, 5)
        self.assertAlmostEqual(15.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[2]
        self.assertAlmostEqual(15.0, theta1, 5)
        self.assertAlmostEqual(25.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[3]
        self.assertAlmostEqual(25.0, theta1, 5)
        self.assertAlmostEqual(35.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[4]
        self.assertAlmostEqual(35.0, theta1, 5)
        self.assertAlmostEqual(45.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[5]
        self.assertAlmostEqual(45.0, theta1, 5)
        self.assertAlmostEqual(55.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[6]
        self.assertAlmostEqual(55.0, theta1, 5)
        self.assertAlmostEqual(65.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[7]
        self.assertAlmostEqual(65.0, theta1, 5)
        self.assertAlmostEqual(75.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[8]
        self.assertAlmostEqual(75.0, theta1, 5)
        self.assertAlmostEqual(85.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        xl, xu, ndbx = input.GRIDX.get()
        self.assertAlmostEqual(-4e-5, xl, 5)
        self.assertAlmostEqual(4e-5, xu, 5)
        self.assertEqual(60, ndbx)

        yl, yu, ndby = input.GRIDY.get()
        self.assertAlmostEqual(-4e-5, yl, 5)
        self.assertAlmostEqual(4e-5, yu, 5)
        self.assertEqual(60, ndby)

        zl, zu, ndbz = input.GRIDZ.get()
        self.assertAlmostEqual(-6e-5, zl, 5)
        self.assertAlmostEqual(0.0, zu, 5)
        self.assertEqual(60, ndbz)

        xrlines, = input.XRLINE.get()
        self.assertEqual(1, len(xrlines))

        izs1s200, = xrlines[0]
        self.assertEqual(29010300, izs1s200)

        filename, = input.RESUME.get()
        self.assertEqual('dump1.dat', filename)

        filename, = input.DUMPTO.get()
        self.assertEqual('dump1.dat', filename)

        dumpp, = input.DUMPP.get()
        self.assertAlmostEqual(60.0, dumpp, 5)

        seed1, seed2 = input.RSEED.get()
        self.assertEqual(-10, seed1)
        self.assertEqual(1, seed2)

        izs1s200, idet, tol = input.REFLIN.get()
        self.assertEqual(29010300, izs1s200)
        self.assertEqual(1, idet)
        self.assertAlmostEqual(1.25e-3, tol, 5)

        dshn, = input.NSIMSH.get()
        self.assertAlmostEqual(2e9, dshn, 5)

        timea, = input.TIME.get()
        self.assertAlmostEqual(2e9, timea, 5)

    def test_epma1_skeleton(self):
        input = create_epma1()
        self._test_epma1(input)

    def test_epma1_write(self):
        input = create_epma1()
        input = self._write_read_input(input)
        self._test_epma1(input)

    def test_epma1_read(self):
        filepath = os.path.join(self.testdatadir, 'epma1.in')
        input = PenepmaInput()
        with open(filepath, 'r') as fp:
            input.read(fp)
        self._test_epma1(input)

    def _test_epma2(self, input):
        se0, = input.SENERG.get()
        self.assertAlmostEqual(15e3, se0, 5)

        sx0, sy0, sz0 = input.SPOSIT.get()
        self.assertAlmostEqual(2e-5, sx0, 8)
        self.assertAlmostEqual(0.0, sy0, 5)
        self.assertAlmostEqual(1.0, sz0, 5)

        theta, phi = input.SDIREC.get()
        self.assertAlmostEqual(180.0, theta, 5)
        self.assertAlmostEqual(0.0, phi, 5)

        alpha, = input.SAPERT.get()
        self.assertAlmostEqual(0.0, alpha, 5)

        materials, = input.materials.get()
        self.assertEqual(2, len(materials))

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[0]
        self.assertEqual('Cu.mat', filename)
        self.assertAlmostEqual(1e3, eabs1, 5)
        self.assertAlmostEqual(1e3, eabs2, 5)
        self.assertAlmostEqual(1e3, eabs3, 5)
        self.assertAlmostEqual(0.2, c1, 5)
        self.assertAlmostEqual(0.2, c2, 5)
        self.assertAlmostEqual(1e3, wcc, 5)
        self.assertAlmostEqual(1e3, wcr, 5)

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[1]
        self.assertEqual('Fe.mat', filename)
        self.assertAlmostEqual(1e3, eabs1, 5)
        self.assertAlmostEqual(1e3, eabs2, 5)
        self.assertAlmostEqual(1e3, eabs3, 5)
        self.assertAlmostEqual(0.2, c1, 5)
        self.assertAlmostEqual(0.2, c2, 5)
        self.assertAlmostEqual(1e3, wcc, 5)
        self.assertAlmostEqual(1e3, wcr, 5)

        self.assertEqual('epma2.geo', input.GEOMFN.get()[0])

        dsmaxs, = input.DSMAX.get()
        self.assertEqual(2, len(dsmaxs))

        kb, dsmax = dsmaxs[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(1e-4, dsmax, 5)

        kb, dsmax = dsmaxs[1]
        self.assertEqual(2, kb)
        self.assertAlmostEqual(1e-4, dsmax, 5)

        iforces, = input.IFORCE.get()
        self.assertEqual(8, len(iforces))

        kb, kpar, icol, forcer, wlow, whig = iforces[0]
        self.assertEqual(1, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(4, icol)
        self.assertAlmostEqual(-5, forcer, 5)
        self.assertAlmostEqual(0.9, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[1]
        self.assertEqual(1, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(5, icol)
        self.assertAlmostEqual(-250, forcer, 5)
        self.assertAlmostEqual(0.9, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[2]
        self.assertEqual(1, kb)
        self.assertEqual(2, kpar)
        self.assertEqual(2, icol)
        self.assertAlmostEqual(10, forcer, 5)
        self.assertAlmostEqual(1e-3, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[3]
        self.assertEqual(1, kb)
        self.assertEqual(2, kpar)
        self.assertEqual(3, icol)
        self.assertAlmostEqual(10, forcer, 5)
        self.assertAlmostEqual(1e-3, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[4]
        self.assertEqual(2, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(4, icol)
        self.assertAlmostEqual(-5, forcer, 5)
        self.assertAlmostEqual(0.9, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[5]
        self.assertEqual(2, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(5, icol)
        self.assertAlmostEqual(-7, forcer, 5)
        self.assertAlmostEqual(0.9, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[6]
        self.assertEqual(2, kb)
        self.assertEqual(2, kpar)
        self.assertEqual(2, icol)
        self.assertAlmostEqual(10, forcer, 5)
        self.assertAlmostEqual(1e-3, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[7]
        self.assertEqual(2, kb)
        self.assertEqual(2, kpar)
        self.assertEqual(3, icol)
        self.assertAlmostEqual(10, forcer, 5)
        self.assertAlmostEqual(1e-3, wlow, 5)
        self.assertAlmostEqual(1.0, whig, 5)

        ibrspls, = input.IBRSPL.get()
        self.assertEqual(2, len(ibrspls))

        kb, factor = ibrspls[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        kb, factor = ibrspls[1]
        self.assertEqual(2, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        ixrspls, = input.IXRSPL.get()
        self.assertEqual(2, len(ixrspls))

        kb, factor = ixrspls[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        kb, factor = ixrspls[1]
        self.assertEqual(2, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        el, eu, nbe = input.NBE.get()
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(300, nbe)

        nbth, nbph = input.NBANGL.get()
        self.assertEqual(45, nbth)
        self.assertEqual(30, nbph)

        photon_detectors, = input.photon_detectors.get()
        self.assertEqual(9, len(photon_detectors))

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[0]
        self.assertAlmostEqual(0.0, theta1, 5)
        self.assertAlmostEqual(90.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[1]
        self.assertAlmostEqual(5.0, theta1, 5)
        self.assertAlmostEqual(15.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[2]
        self.assertAlmostEqual(15.0, theta1, 5)
        self.assertAlmostEqual(25.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[3]
        self.assertAlmostEqual(25.0, theta1, 5)
        self.assertAlmostEqual(35.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[4]
        self.assertAlmostEqual(35.0, theta1, 5)
        self.assertAlmostEqual(45.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[5]
        self.assertAlmostEqual(45.0, theta1, 5)
        self.assertAlmostEqual(55.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[6]
        self.assertAlmostEqual(55.0, theta1, 5)
        self.assertAlmostEqual(65.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[7]
        self.assertAlmostEqual(65.0, theta1, 5)
        self.assertAlmostEqual(75.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        (theta1, theta2, phi1, phi2, ipsf,
         edel, edeu, nche, emission_filename) = photon_detectors[8]
        self.assertAlmostEqual(75.0, theta1, 5)
        self.assertAlmostEqual(85.0, theta2, 5)
        self.assertAlmostEqual(0.0, phi1, 5)
        self.assertAlmostEqual(360.0, phi2, 5)
        self.assertEqual(0, ipsf)
        self.assertAlmostEqual(0.0, edel, 5)
        self.assertAlmostEqual(0.0, edeu, 5)
        self.assertEqual(1000, nche)
        self.assertIsNone(emission_filename)

        xl, xu, ndbx = input.GRIDX.get()
        self.assertAlmostEqual(-1e-5, xl, 5)
        self.assertAlmostEqual(5e-5, xu, 5)
        self.assertEqual(60, ndbx)

        yl, yu, ndby = input.GRIDY.get()
        self.assertAlmostEqual(-3e-5, yl, 5)
        self.assertAlmostEqual(3e-5, yu, 5)
        self.assertEqual(60, ndby)

        zl, zu, ndbz = input.GRIDZ.get()
        self.assertAlmostEqual(-6e-5, zl, 5)
        self.assertAlmostEqual(0.0, zu, 5)
        self.assertEqual(60, ndbz)

        xrlines, = input.XRLINE.get()
        self.assertEqual(2, len(xrlines))

        izs1s200, = xrlines[0]
        self.assertEqual(26010300, izs1s200)

        izs1s200, = xrlines[1]
        self.assertEqual(29010300, izs1s200)

        filename, = input.RESUME.get()
        self.assertEqual('dump2.dat', filename)

        filename, = input.DUMPTO.get()
        self.assertEqual('dump2.dat', filename)

        dumpp, = input.DUMPP.get()
        self.assertAlmostEqual(60.0, dumpp, 5)

        seed1, seed2 = input.RSEED.get()
        self.assertEqual(-10, seed1)
        self.assertEqual(1, seed2)

        izs1s200, idet, tol = input.REFLIN.get()
        self.assertEqual(26010300, izs1s200)
        self.assertEqual(1, idet)
        self.assertAlmostEqual(1.5e-3, tol, 5)

        dshn, = input.NSIMSH.get()
        self.assertAlmostEqual(2e9, dshn, 5)

        timea, = input.TIME.get()
        self.assertAlmostEqual(2e9, timea, 5)

    def test_epma2_skeleton(self):
        input = create_epma2()
        self._test_epma2(input)

    def test_epma2_write(self):
        input = create_epma2()
        input = self._write_read_input(input)
        self._test_epma2(input)

    def test_epma2_read(self):
        filepath = os.path.join(self.testdatadir, 'epma2.in')
        input = PenepmaInput()
        with open(filepath, 'r') as fp:
            input.read(fp)
        self._test_epma2(input)

if __name__ == '__main__': #pragma: no cover
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import os
import io

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.transformation import Rotation, Shift, Scale
from pypenelopetools.pengeom.base import LINE_SEPARATOR

# Globals and constants variables.

class TestRotation(unittest.TestCase):

    LINES = ['  OMEGA=(+1.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
             '  THETA=(+2.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
             '    PHI=(+3.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
             '0000000000000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        self.rotation = Rotation(1.0, 2.0, 3.0)

    def _test_rotation(self, rotation):
        self.assertAlmostEqual(1.0, rotation.omega_deg, 4)
        self.assertAlmostEqual(2.0, rotation.theta_deg, 4)
        self.assertAlmostEqual(3.0, rotation.phi_deg, 4)

    def testskeleton(self):
        self._test_rotation(self.rotation)

    def test_write_read(self):
        fileobj = io.StringIO()

        try:
            self.rotation._write(fileobj, {})
            fileobj.write(LINE_SEPARATOR + os.linesep)

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES, lines)

            fileobj.seek(0)
            rotation = Rotation()
            rotation._read(fileobj, {}, {}, {})
            self._test_rotation(rotation)
        finally:
            fileobj.close()

class TestShift(unittest.TestCase):

    LINES = ['X-SHIFT=(+1.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+2.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+3.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        self.shift = Shift(1.0, 2.0, 3.0)

    def _test_shift(self, shift):
        self.assertAlmostEqual(1.0, shift.x_cm, 4)
        self.assertAlmostEqual(2.0, shift.y_cm, 4)
        self.assertAlmostEqual(3.0, shift.z_cm, 4)

    def testskeleton(self):
        self._test_shift(self.shift)

    def test_write_read(self):
        fileobj = io.StringIO()

        try:
            self.shift._write(fileobj, {})
            fileobj.write(LINE_SEPARATOR + os.linesep)

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES, lines)

            fileobj.seek(0)
            shift = Shift()
            shift._read(fileobj, {}, {}, {})
            self._test_shift(shift)
        finally:
            fileobj.close()

class TestScale(unittest.TestCase):

    LINES = ['X-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Y-SCALE=(+2.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Z-SCALE=(+3.000000000000000E+00,   0)              (DEFAULT=1.0)',
               '0000000000000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        self.scale = Scale(1.0, 2.0, 3.0)

    def _test_scale(self, scale):
        self.assertAlmostEqual(1.0, scale.x, 4)
        self.assertAlmostEqual(2.0, scale.y, 4)
        self.assertAlmostEqual(3.0, scale.z, 4)

    def testskeleton(self):
        self._test_scale(self.scale)

    def test_write_read(self):
        fileobj = io.StringIO()

        try:
            self.scale._write(fileobj, {})
            fileobj.write(LINE_SEPARATOR + os.linesep)

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES, lines)

            fileobj.seek(0)
            scale = Scale()
            scale._read(fileobj, {}, {}, {})
            self._test_scale(scale)
        finally:
            fileobj.close()

if __name__ == '__main__': #pragma: no cover
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import io
import os

# Third party modules.

# Local modules.
from pypenelopetools.penmain.input import PenmainInput
from pypenelopetools.material import Material
from pypenelopetools.pengeom.surface import sphere, zplane, cylinder
from pypenelopetools.pengeom.module import Module, SidePointer
from pypenelopetools.pengeom.geometry import Geometry

# Globals and constants variables.

def create_example1_disc():
    # Create materials
    material_cu = Material('Cu', {29: 1.0}, 8.9)

    # Create geometry
    module = Module(material_cu, 'Solid cylinder')
    module.add_surface(zplane(0.0), SidePointer.POSITIVE)
    module.add_surface(zplane(0.005), SidePointer.NEGATIVE)
    module.add_surface(cylinder(1.0), SidePointer.NEGATIVE)

    geometry = Geometry('A solid cylinder.')
    geometry.add_module(module)

    # Create input
    input = PenmainInput()

    input.TITLE.set('Point source and a homogeneous cylinder.')
    input.SKPAR.set(1)
    input.SENERG.set(40e3)
    input.SPOSIT.set(0.0, 0.0, -0.0001)
    input.SCONE.set(0.0, 0.0, 5.0)

    input.materials.add(1, material_cu.filename, 1e3, 1e3, 1e3, 0.05, 0.05, 1e3, 1e3)

    input.GEOMFN.set('disc.geo')
    input.PARINP.add(1, 0.005)
    input.PARINP.add(2, 0.01)
    input.DSMAX.add(1, 1e-4)

    input.IFORCE.add(1, 1, 4, 2000, 0.1, 2.0)
    input.IFORCE.add(1, 1, 5, 200, 0.1, 2.0)
    input.IBRSPL.add(1, 2)
    input.IXRSPL.add(1, 2)

    input.NBE.set(0, 0, 100)
    input.NBANGL.set(45, 18)

    detector = input.impact_detectors.add(0.0, 0.0, 100, 0, 2)
    detector.IDBODY.add(1)

    detector = input.energy_deposition_detectors.add(0, 0, 100)
    detector.EDBODY.add(1)

    input.GRIDZ.set(0, 0.005, 100)
    input.GRIDR.set(0.01, 50)

    input.RESUME.set('dump.dat')
    input.DUMPTO.set('dump.dat')
    input.DUMPP.set(60)

    input.NSIMSH.set(2e9)
    input.TIME.set(600)

    return input

def create_example2_plane():
    # Create materials
    material_h2o = Material('H2O', {8: 0.8881, 1: 0.1119}, 1.0)

    # Create geometry
    module_detector = Module(material_h2o, 'Fluence detector')
    module_detector.add_surface(sphere(2.0), SidePointer.NEGATIVE)

    module_phantom = Module(material_h2o, 'Water phantom')
    module_phantom.add_surface(zplane(0.0), SidePointer.POSITIVE)
    module_phantom.add_surface(zplane(30.0), SidePointer.NEGATIVE)

    geometry = Geometry('Semi-infinite water phantom')
    geometry.add_module(module_detector)
    geometry.add_module(module_phantom)

    index_lookup = geometry.indexify()

    # Create input
    input = PenmainInput()

    input.TITLE.set('Dose in a water phantom with a spherical impact detector')
    input.SKPAR.set(2)
    input.SENERG.set(3e7)
    input.SPOSIT.set(0.0, 0.0, -25.0)
    input.SCONE.set(0.0, 0.0, 5.0)

    input.materials.add(1, material_h2o.filename, 1e5, 1e4, 1e5, 0.05, 0.05, 5e3, 5e3)

    input.GEOMFN.set('plane.geo')

    input.NBE.set(1e5, 3e7, 100)
    input.NBANGL.set(45, 18)

    detector = input.impact_detectors.add(1e5, 0.0, 100, 0, 2)
    detector.IDBODY.add(index_lookup[module_detector])

    input.GRIDZ.set(0, 30.0, 60)
    input.GRIDR.set(30.0, 60.0)

    input.RESUME.set('dump.dat')
    input.DUMPTO.set('dump.dat')
    input.DUMPP.set(60)

    input.NSIMSH.set(1e7)
    input.TIME.set(2e9)

    return input

class TestPenmainInput(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)

        self.testdatadir = os.path.join(os.path.dirname(__file__),
                                        '..', 'testdata', 'penmain')

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def _write_read_input(self, input):
        fileobj = io.StringIO()

        try:
            input.write(fileobj)

            fileobj.seek(0)
            outinput = PenmainInput()
            outinput.read(fileobj)
        finally:
            fileobj.close()

        return outinput

    def _test_example1_disc(self, input):
        kparp, = input.SKPAR.get()
        self.assertEqual(1, kparp)

        se0, = input.SENERG.get()
        self.assertAlmostEqual(40e3, se0, 5)

        sx0, sy0, sz0 = input.SPOSIT.get()
        self.assertAlmostEqual(0.0, sx0, 5)
        self.assertAlmostEqual(0.0, sy0, 5)
        self.assertAlmostEqual(-0.0001, sz0, 5)

        theta, phi, alpha = input.SCONE.get()
        self.assertAlmostEqual(0.0, theta, 5)
        self.assertAlmostEqual(0.0, phi, 5)
        self.assertAlmostEqual(5.0, alpha, 5)

        materials, = input.materials.get()
        self.assertEqual(1, len(materials))

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[0]
        self.assertEqual('Cu.mat', filename)
        self.assertAlmostEqual(1e3, eabs1, 5)
        self.assertAlmostEqual(1e3, eabs2, 5)
        self.assertAlmostEqual(1e3, eabs3, 5)
        self.assertAlmostEqual(0.05, c1, 5)
        self.assertAlmostEqual(0.05, c2, 5)
        self.assertAlmostEqual(1e3, wcc, 5)
        self.assertAlmostEqual(1e3, wcr, 5)

        self.assertEqual('disc.geo', input.GEOMFN.get()[0])

        parinps, = input.PARINP.get()
        self.assertEqual(2, len(parinps))

        ip, parinp = parinps[0]
        self.assertEqual(1, ip)
        self.assertAlmostEqual(0.005, parinp, 5)

        ip, parinp = parinps[1]
        self.assertEqual(2, ip)
        self.assertAlmostEqual(0.01, parinp, 5)

        dsmaxs, = input.DSMAX.get()
        self.assertEqual(1, len(dsmaxs))

        kb, dsmax = dsmaxs[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(1e-4, dsmax, 5)

        iforces, = input.IFORCE.get()
        self.assertEqual(2, len(iforces))

        kb, kpar, icol, forcer, wlow, whig = iforces[0]
        self.assertEqual(1, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(4, icol)
        self.assertAlmostEqual(2000, forcer, 5)
        self.assertAlmostEqual(0.1, wlow, 5)
        self.assertAlmostEqual(2.0, whig, 5)

        kb, kpar, icol, forcer, wlow, whig = iforces[1]
        self.assertEqual(1, kb)
        self.assertEqual(1, kpar)
        self.assertEqual(5, icol)
        self.assertAlmostEqual(200, forcer, 5)
        self.assertAlmostEqual(0.1, wlow, 5)
        self.assertAlmostEqual(2.0, whig, 5)

        ibrspls, = input.IBRSPL.get()
        self.assertEqual(1, len(ibrspls))

        kb, factor = ibrspls[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        ixrspls, = input.IXRSPL.get()
        self.assertEqual(1, len(ixrspls))

        kb, factor = ixrspls[0]
        self.assertEqual(1, kb)
        self.assertAlmostEqual(2.0, factor, 5)

        el, eu, nbe = input.NBE.get()
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(100, nbe)

        nbth, nbph = input.NBANGL.get()
        self.assertEqual(45, nbth)
        self.assertEqual(18, nbph)

        impact_detectors, = input.impact_detectors.get()
        self.assertEqual(1, len(impact_detectors))

        (el, eu, nbe, ipsf, idcut,
         spectrum_filename, psf_filename, fln_filename,
         agel, ageu, nage, age_filename, kbs, kpars) = impact_detectors[0]
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(100, nbe)
        self.assertEqual(0, ipsf)
        self.assertEqual(2, idcut)
        self.assertIsNone(spectrum_filename)
        self.assertIsNone(psf_filename)
        self.assertIsNone(fln_filename)
        self.assertIsNone(agel)
        self.assertIsNone(ageu)
        self.assertIsNone(nage)
        self.assertIsNone(age_filename)
        self.assertEqual(1, len(kbs))
        self.assertEqual(0, len(kpars))

        energy_deposition_detectors, = input.energy_deposition_detectors.get()
        self.assertEqual(1, len(energy_deposition_detectors))

        el, eu, nbe, spectrum_filename, kbs = energy_deposition_detectors[0]
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(100, nbe)
        self.assertIsNone(spectrum_filename)
        self.assertEqual(1, len(kbs))

        zl, zu, ndbz = input.GRIDZ.get()
        self.assertAlmostEqual(0.0, zl, 5)
        self.assertAlmostEqual(0.005, zu, 5)
        self.assertEqual(100, ndbz)

        ru, ndbr = input.GRIDR.get()
        self.assertAlmostEqual(0.01, ru, 5)
        self.assertEqual(50, ndbr)

        filename, = input.RESUME.get()
        self.assertEqual('dump.dat', filename)

        filename, = input.DUMPTO.get()
        self.assertEqual('dump.dat', filename)

        dumpp, = input.DUMPP.get()
        self.assertAlmostEqual(60.0, dumpp, 5)

        dshn, = input.NSIMSH.get()
        self.assertAlmostEqual(2e9, dshn, 5)

        timea, = input.TIME.get()
        self.assertAlmostEqual(600.0, timea, 5)

    def test_example1_disc_write(self):
        input = create_example1_disc()
        input = self._write_read_input(input)
        self._test_example1_disc(input)

    def test_example1_disc_read(self):
        filepath = os.path.join(self.testdatadir, '1-disc', 'disc.in')
        input = PenmainInput()
        with open(filepath, 'r') as fp:
            input.read(fp)
        self._test_example1_disc(input)

    def _test_example2_plane(self, input):
        kparp, = input.SKPAR.get()
        self.assertEqual(2, kparp)

        se0, = input.SENERG.get()
        self.assertAlmostEqual(3e7, se0, 5)

        sx0, sy0, sz0 = input.SPOSIT.get()
        self.assertAlmostEqual(0.0, sx0, 5)
        self.assertAlmostEqual(0.0, sy0, 5)
        self.assertAlmostEqual(-25.0, sz0, 5)

        theta, phi, alpha = input.SCONE.get()
        self.assertAlmostEqual(0.0, theta, 5)
        self.assertAlmostEqual(0.0, phi, 5)
        self.assertAlmostEqual(5.0, alpha, 5)

        materials, = input.materials.get()
        self.assertEqual(1, len(materials))

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[0]
        self.assertEqual('H2O.mat', filename)
        self.assertAlmostEqual(1e5, eabs1, 5)
        self.assertAlmostEqual(1e4, eabs2, 5)
        self.assertAlmostEqual(1e5, eabs3, 5)
        self.assertAlmostEqual(0.05, c1, 5)
        self.assertAlmostEqual(0.05, c2, 5)
        self.assertAlmostEqual(5e3, wcc, 5)
        self.assertAlmostEqual(5e3, wcr, 5)

        self.assertEqual('plane.geo', input.GEOMFN.get()[0])

        el, eu, nbe = input.NBE.get()
        self.assertAlmostEqual(1e5, el, 5)
        self.assertAlmostEqual(3e7, eu, 5)
        self.assertEqual(100, nbe)

        nbth, nbph = input.NBANGL.get()
        self.assertEqual(45, nbth)
        self.assertEqual(18, nbph)

        impact_detectors, = input.impact_detectors.get()
        self.assertEqual(1, len(impact_detectors))

        (el, eu, nbe, ipsf, idcut,
         spectrum_filename, psf_filename, fln_filename,
         agel, ageu, nage, age_filename, kbs, kpars) = impact_detectors[0]
        self.assertAlmostEqual(1e5, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(100, nbe)
        self.assertEqual(0, ipsf)
        self.assertEqual(2, idcut)
        self.assertIsNone(spectrum_filename)
        self.assertIsNone(psf_filename)
        self.assertIsNone(fln_filename)
        self.assertIsNone(agel)
        self.assertIsNone(ageu)
        self.assertIsNone(nage)
        self.assertIsNone(age_filename)
        self.assertEqual(1, len(kbs))
        self.assertEqual(0, len(kpars))

        zl, zu, ndbz = input.GRIDZ.get()
        self.assertAlmostEqual(0.0, zl, 5)
        self.assertAlmostEqual(30.0, zu, 5)
        self.assertEqual(60, ndbz)

        ru, ndbr = input.GRIDR.get()
        self.assertAlmostEqual(30.0, ru, 5)
        self.assertEqual(60, ndbr)

        filename, = input.RESUME.get()
        self.assertEqual('dump.dat', filename)

        filename, = input.DUMPTO.get()
        self.assertEqual('dump.dat', filename)

        dumpp, = input.DUMPP.get()
        self.assertAlmostEqual(60.0, dumpp, 5)

        dshn, = input.NSIMSH.get()
        self.assertAlmostEqual(1e7, dshn, 5)

        timea, = input.TIME.get()
        self.assertAlmostEqual(2e9, timea, 5)

    def test_example2_plane_write(self):
        input = create_example2_plane()
        input = self._write_read_input(input)
        self._test_example2_plane(input)

    def test_example2_plane_read(self):
        filepath = os.path.join(self.testdatadir, '2-plane', 'plane.in')
        input = PenmainInput()
        with open(filepath, 'r') as fp:
            input.read(fp)
        self._test_example2_plane(input)

if __name__ == '__main__': #pragma: no cover
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



"""
Definition of base keyword classes.
"""

# Standard library modules.
import abc
import os

# Third party modules.

# Local modules.
from pypenelopetools.penelope.base import _InputLineBase

# Globals and constants variables.

#--- Abstract classes

class _KeywordBase(_InputLineBase):
    """
    Base of all PENELOPE keywords.
    """

    def __str__(self):
        return self.name

    @abc.abstractmethod
    def get(self):
        """
        Returns:
            tuple: Value(s) of this keyword.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self):
        """
        Returns:
            _KeyboardBase: Deep copy of this keyword.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def name(self):
        """str: Name of keyword."""
        raise NotImplementedError

#--- Core classes

class TypeKeyword(_KeywordBase):
    """
    Keyword where the :obj:`type` of the values are checked.
    
    Args:
        name (str): Name of keyword.
        types (tuple(type)): Type of each value of the keyword.
        comment (str, optional): Comment.
    """

    def __init__(self, name, types, comment=""):
        self._name = name
        self._types = tuple(types)
        self._values = tuple([None] * len(types))
        self._comment = comment

    def set(self, *args):
        """
        Sets the value(s) of the keyword. 
        Each value is checked if it matches the defined type.
        
        Args:
            *args: Value(s).
            
        Raises:
            TypeError: If one value does not match its defined type.
        """
        if len(args) < len(self._types): # Less than to account for additional, not parsed values
            raise ValueError("Keyword {0} requires {1} values, {2} given"
                             .format(self.name, len(self._types), len(args)))

        values = []
        for type_, value in zip(self._types, args):
            if value is not None:
                try:
                    value = type_(value)
                except ValueError:
                    raise TypeError("Value {0} must be of type {1}"
                                    .format(value, type_))
            values.append(value)

        self._values = tuple(values)

    def get(self):
        return self._values

    def copy(self):
        return self.__class__(self.name, self._types, self.comment)

    def read(self, fileobj):
        line = self._peek_next_line(fileobj)
        name, values, _comment = self._parse_line(line)

        # If it is not the expected line, do nothing
        if name != self.name:
            return

        # Set values
        self.set(*values)

        # Jump to next line
        self._read_next_line(fileobj)

    def write(self, fileobj):
        values = list(self.get())

        # Skip if no values are defined
        if None in values:
            return

        # Write to file
        line = self._create_line(self.name, values, self.comment)
        fileobj.write(line + os.linesep)

    @property
    def name(self):
        return self._name

    @property
    def comment(self):
        """str: Comment."""
        return self._comment

class KeywordGroup(_KeywordBase):
    """
    Group of keywords, keywords that should always be defined together.
    """

    @abc.abstractmethod
    def get_keywords(self):
        """
        Returns:
            tuple: Keywords apart of this group.
        """
        raise NotImplementedError

    def get(self):
        """
        Returns:
            tuple: Value(s) of all keywords.
        """
        values = []

        for keyword in self.get_keywords():
            values.extend(keyword.get())

        return tuple(values)

    def copy(self):
        return self.__class__()

    def read(self, fileobj):
        for keyword in self.get_keywords():
            keyword.read(fileobj)

    def write(self, fileobj):
        for keyword in self.get_keywords():
            keyword.write(fileobj)

    def _set_keyword_sequence(self, keyword, values):
        keyword.clear()
        if values is None:
            return

        if not isinstance(values, (list, tuple)):
            values = [values]

        for args in values:
            keyword.add(*args)

    @property
    def name(self):
        return self.get_keywords()[0].name

class KeywordSequence(_KeywordBase):
    """
    Sequence of keywords, keywords that can be defined multiple times.
    
    Args:
        keyword (_KeywordBase): Base keyword.
        maxlength (int): Maximum number of keywords that can be added.
    """

    def __init__(self, keyword, maxlength):
        self._base_keyword = keyword
        self._keywords = []
        self._maxlength = maxlength

    def _create_keyword(self):
        return self._base_keyword.copy()

    def _add_keyword(self, keyword):
        if len(self._keywords) >= self._maxlength:
            raise ValueError('Exceeded maximum number of keywords.')
        self._keywords.append(keyword)

    def add(self, *args):
        """Adds a new keyword definition.
        This internally creates a new keyword based on the base keyword,
        sets the value(s) and add it to a list.
        
        Args:
            *args: Value(s).
        """
        keyword = self._create_keyword()
        keyword.set(*args)
        self._add_keyword(keyword)
        return keyword

    def pop(self, index):
        """Removes a keyword.
        
        Args:
            index (int): Index of the keyword to be removed.
        """
        self._keywords.pop(index)

    def clear(self):
        """Clears all added keywords."""
        self._keywords.clear()

    def get(self):
        """
        Returns:
            tuple: Value(s) of all keywords.
        """
        values = []
        for keyword in self._keywords:
            values.append(keyword.get())
        return (tuple(values),)

    def copy(self):
        return self.__class__(self._base_keyword.copy())

    def read(self, fileobj):
        line = self._peek_next_line(fileobj)
        name, _values, _comment = self._parse_line(line)

        while name == self._base_keyword.name:
            # Read keyword
            keyword = self._create_keyword()
            keyword.read(fileobj)
            self._add_keyword(keyword)

            # Try to read line
            line = self._peek_next_line(fileobj)
            name, _values, _comment = self._parse_line(line)

    def write(self, fileobj):
        for keyword in self._keywords:
            keyword.write(fileobj)

    @property
    def name(self):
        return self._base_keyword.name




"""
Definition of base input classes.
"""

# Standard library modules.
import abc

# Third party modules.

# Local modules.
from pypenelopetools.penelope.base import _InputLineBase

# Globals and constants variables.

class _PenelopeInputBase(_InputLineBase):
    """
    Base input class.
    """

    def read(self, fileobj):
        """
        Reads an input file (i.e. ``.in``).
        
        Args:
            fileobj (file object): File object opened with read access.
        """
        for keyword in self.get_keywords():
            keyword.read(fileobj)

    def write(self, fileobj):
        """
        Writes to an input file (i.e. ``.in``).
        
        Args:
            fileobj (file object): File object opened with write access.
        """
        for keyword in self.get_keywords():
            keyword.write(fileobj)

    @abc.abstractmethod
    def get_keywords(self):
        """
        Returns:
            list: Sorted list of all keywords in the input.
        """
        return []



"""
Mixins used to create geometry objects.
"""

# Standard library modules.

# Third party modules.

# Local modules.

# Globals and constants variables.

class DescriptionMixin(object):
    """
    Mixin that adds a description property.
    """

    @property
    def description(self):
        """str: Description of the geometry object."""
        return self._description

    @description.setter
    def description(self, desc):
        self._description = desc

class ModuleMixin(object):
    """
    Mixin that adds methods to add, pop and clear modules.
    """

    def add_module(self, module):
        """
        Adds a module.
        
        Args:
            module (:obj:`Module <pypenelopetools.pengeom.module.Module>`):
                Module to add.
        """
        if module == self:
            raise ValueError("Cannot add this module to this module.")
        self._modules.add(module)

    def pop_module(self, module):
        """
        Removes a module.
        
        Args:
            module (:obj:`Module <pypenelopetools.pengeom.module.Module>`):
                Module to remove.
        """
        self._modules.discard(module)

    def clear_modules(self):
        """
        Clear all modules.
        """
        self._modules.clear()

    def get_modules(self):
        """
        Returns:
            tuple: All modules.
        """
        return tuple(self._modules)


""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.input import _PenelopeInputBase
import pypenelopetools.penelope.keywords as penelope_keywords
import pypenelopetools.penmain.keywords as penmain_keywords
import pypenelopetools.penelope.separators as penelope_separators
import pypenelopetools.penmain.separators as penmain_separators

# Globals and constants variables.

class PenmainInput(_PenelopeInputBase):

    def __init__(self):
        self.TITLE = penelope_keywords.TITLE()

        # Source definition
        self.SKPAR = penelope_keywords.SKPAR()
        self.SENERG = penelope_keywords.SENERG()
        self.SPECTR = penelope_keywords.SPECTR()
        self.SGPOL = penelope_keywords.SGPOL()
        self.SPOSIT = penelope_keywords.SPOSIT()
        self.SBOX = penmain_keywords.SBOX()
        self.SBODY = penmain_keywords.SBODY()
        self.SCONE = penelope_keywords.SCONE()
        self.SRECTA = penelope_keywords.SRECTA()

        # Input phase-space file
        self.IPSFN = penmain_keywords.IPSFN()
        self.IPSPLI = penmain_keywords.IPSPLI()
        self.WGTWIN = penmain_keywords.WGTWIN()
        self.EPMAX = penmain_keywords.EPMAX()

        # Material data and simulation parameters
        self.materials = penelope_keywords.Materials()

        # Geometry and local simulation parameters
        self.GEOMFN = penelope_keywords.GEOMFN()
        self.PARINP = penmain_keywords.PARINP()
        self.DSMAX = penelope_keywords.DSMAX()
        self.EABSB = penelope_keywords.EABSB()

        # Interaction forcing
        self.IFORCE = penelope_keywords.IFORCE()

        # Bremsstrahlung splitting
        self.IBRSPL = penelope_keywords.IBRSPL()

        # X-ray splitting
        self.IXRSPL = penelope_keywords.IXRSPL()

        # Emerging particles
        self.NBE = penelope_keywords.NBE()
        self.NBANGL = penelope_keywords.NBANGL()

        # Impact detectors
        self.impact_detectors = penmain_keywords.ImpactDetectors()

        # Energy-deposition detectors
        self.energy_deposition_detectors = penmain_keywords.EnergyDepositionDetectors()

        # Absorbed dose distribution
        self.GRIDX = penelope_keywords.GRIDX()
        self.GRIDY = penelope_keywords.GRIDY()
        self.GRIDZ = penelope_keywords.GRIDZ()
        self.GRIDR = penmain_keywords.GRIDR()

        # Job properties
        self.RESUME = penelope_keywords.RESUME()
        self.DUMPTO = penelope_keywords.DUMPTO()
        self.DUMPP = penelope_keywords.DUMPP()
        self.RSEED = penelope_keywords.RSEED()
        self.NSIMSH = penelope_keywords.NSIMSH()
        self.TIME = penelope_keywords.TIME()

    def get_keywords(self):
        return [self.TITLE,

                penelope_separators.DOT,
                penelope_separators.SOURCE_DEFINITION,
                self.SKPAR, self.SENERG, self.SPECTR, self.SGPOL,
                self.SPOSIT, self.SBOX, self.SBODY, self.SCONE, self.SRECTA,

                penelope_separators.DOT,
                penmain_separators.INPUT_PHASE_SPACE_FILE,
                self.IPSFN, self.IPSPLI, self.WGTWIN, self.EPMAX,

                penelope_separators.DOT,
                penelope_separators.MATERIAL,
                self.materials,

                penelope_separators.DOT,
                penmain_separators.GEOMETRY,
                self.GEOMFN, self.PARINP, self.DSMAX, self.EABSB,

                penelope_separators.DOT,
                penelope_separators.INTERACTION_FORCING,
                self.IFORCE,

                penelope_separators.DOT,
                penelope_separators.BREMSSTRAHLUNG_SPLITTING,
                self.IBRSPL,

                penelope_separators.DOT,
                penelope_separators.XRAY_SPLITTING,
                self.IXRSPL,

                penelope_separators.DOT,
                penelope_separators.EMERGING_PARTICLES,
                self.NBE, self.NBANGL,

                penelope_separators.DOT,
                penmain_separators.IMPACT_DETECTORS,
                self.impact_detectors,

                penelope_separators.DOT,
                penelope_separators.ENERGY_DEPOSITON_DETECTORS,
                self.energy_deposition_detectors,

                penelope_separators.DOT,
                penmain_separators.ABSORBED_DOSE_DISTRIBUTION,
                self.GRIDX, self.GRIDY, self.GRIDZ, self.GRIDR,

                penelope_separators.DOT,
                penelope_separators.JOB_PROPERTIES,
                self.RESUME, self.DUMPTO, self.DUMPP,

                penelope_separators.DOT,
                self.RSEED, self.NSIMSH, self.TIME,

                penelope_separators.END]



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import io

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.surface import SurfaceImplicit, SurfaceReduced

# Globals and constants variables.

class TestSurfaceImplicit(unittest.TestCase):

    LINES = ['SURFACE (   1) surface',
               'INDICES=( 0, 0, 0, 0, 0)',
               '    AXX=(+1.000000000000000E+03,   0)              (DEFAULT=0.0)',
               '    AXY=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '    AXZ=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '    AYY=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '    AYZ=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '    AZZ=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '     AX=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '     AY=(+1.000000000000000E+09,   0)              (DEFAULT=0.0)',
               '     AZ=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '     A0=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '1111111111111111111111111111111111111111111111111111111111111111',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+1.800000000000000E+02,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(-1.000000000000000E+05,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        coefficients = (1e3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1e9, 0.0, 0.0)
        self.surface = SurfaceImplicit(coefficients, description='surface')
        self.surface.rotation.phi_deg = 180
        self.surface.shift.z_cm = -1e5

    def _test_surface(self, surface):
        self.assertEqual('surface', surface.description)
        self.assertAlmostEqual(180, surface.rotation.phi_deg, 4)
        self.assertAlmostEqual(-1e5, surface.shift.z_cm, 4)
        self.assertAlmostEqual(1e3, surface.coefficients['xx'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['xy'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['xz'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['yy'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['yz'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['zz'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['x'], 4)
        self.assertAlmostEqual(1e9, surface.coefficients['y'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['z'], 4)
        self.assertAlmostEqual(0.0, surface.coefficients['0'], 4)

    def testskeleton(self):
        self._test_surface(self.surface)

    def test_write_read(self):
        fileobj = io.StringIO()

        try:
            self.surface._write(fileobj, {self.surface: 1})

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES, lines)

            fileobj.seek(0)
            surface = SurfaceImplicit()
            surface._read(fileobj, {}, {}, {})
            self._test_surface(surface)
        finally:
            fileobj.close()

class TestSurfaceReduced(unittest.TestCase):

    LINES = ['SURFACE (   1) surface',
               'INDICES=( 1, 1, 1, 0,-1)',
               'X-SCALE=(+3.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Y-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Z-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        self.surface = SurfaceReduced((1, 1, 1, 0, -1), 'surface')
        self.surface.scale.x = 3.0

    def _test_surface(self, surface):
        self.assertEqual((1, 1, 1, 0, -1), surface.indices)
        self.assertEqual('surface', surface.description)
        self.assertAlmostEqual(3.0, surface.scale.x, 4)

    def testskeleton(self):
        self._test_surface(self.surface)

    def test_write_read(self):
        fileobj = io.StringIO()

        try:
            self.surface._write(fileobj, {self.surface: 1})

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES, lines)

            fileobj.seek(0)
            surface = SurfaceReduced()
            surface._read(fileobj, {}, {}, {})
            self._test_surface(surface)
        finally:
            fileobj.close()

if __name__ == '__main__': #pragma: no cover
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



""""""

# Standard library modules.
import os

# Third party modules.

# Local modules.
from pypenelopetools.penelope.keyword import \
    TypeKeyword, KeywordGroup, KeywordSequence
import pypenelopetools.penelope.keywords as penelope_keywords

# Globals and constants variables.

class GeometryDefinitionGroup(KeywordGroup):
    """
    This block of lines defines the geometrical structure. The
    only allowed keywords in the geometry definition list are
    'GSTART', 'LAYER_', 'CENTRE', 'CYLIND' and 'GEND__' (notice
    the blanks). The definition list must begin with the
    'GSTART' line and terminate with the 'GEND__' line. The
    second line must be a 'LAYER_' line, which initiates the
    definition of the layer structure. Each 'LAYER_' line is
    followed by a 'CENTRE' line (optional) and by one or several
    'CYLIND' lines, which define the various concentric rings in
    the layer; a layer may be void. No blank lines are allowed
    in the geometry definition list.
    
    Layers must be defined in increasing order of heights, from
    bottom to top of the structure. If the 'CENTRE' line is not
    entered, cylinders are assumed to be centred on the Z-axis
    (XCEN=YCEN=0.0). Cylinders have to be defined in increasing
    radial order, from the centre to the periphery. The two
    lengths in each 'LAYER_' and 'CYLIND' line must be entered
    in increasing order. All numerical data are read in free
    format.
    """

    def __init__(self):
        self.LAYER = TypeKeyword('LAYER', (float, float),
                                 comment='Z_lower and Z_higher')
        self.CENTER = TypeKeyword('CENTER', (float, float),
                                  comment='X_centre and Y_centre')

        keyword = TypeKeyword('CYLIND', (int, float, float),
                              comment='Material, R_inner and R_outer')
        self.CYLIND = KeywordSequence(keyword, maxlength=50)

    def get_keywords(self):
        return (self.LAYER, self.CENTER, self.CYLIND)

    def set(self, zlow, zhigh, xcen=None, ycen=None, cylinders=None):
        self.LAYER.set(zlow, zhigh)
        self.CENTER.set(xcen, ycen)
        self._set_keyword_sequence(self.CYLIND, cylinders)

class GeometryDefinitions(KeywordSequence):

    def __init__(self, maxlength=100):
        keyword = GeometryDefinitionGroup()
        super().__init__(keyword, maxlength)

    def add(self, zlow, zhigh, xcen=None, ycen=None, cylinders=None):
        return super().add(zlow, zhigh, xcen, ycen, cylinders)

class SEXTND(KeywordSequence):
    """
    For internal extended sources, this line defines an active
    body KL,KC (the cylinder KC in layer KL) and its relative
    activity density, RELAC.
      DEFAULT: none.
    
    NOTE: The labels KL,KC that identify each body are defined
    by the ordering in the input geometry list. These labels
    are written in the output geometry report.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword('SEXTND', (int, int, float),
                              comment='Extended source in KL,KC, rel. activity dens.')
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, relac):
        super().add(kl, kc, relac)

class STHICK(TypeKeyword):
    """
    For an external source, thickness (height) of the
    active volume of the source (cylinder or ring).
    DEFAULT: STHICK=0.0
    """

    def __init__(self):
        super().__init__('STHICK', (float,), comment='Source height')

class SRADII(TypeKeyword):
    """
    For an external source, inner and outer radii of the active
    volume (cylinder or ring).
      DEFAULTS: SRIN=0.0, SROUT=0.0
    """

    def __init__(self):
        super().__init__('SRADII', (float, float),
                         comment='Source inner and outer radii')

class DSMAX(KeywordSequence):
    """
    Maximum step length DSMAX(KL,KC) of electrons and positrons
    in cylinder KL,KC. This parameter is important only for thin
    bodies; it should be given a value of the order of one tenth
    of the cylinder thickness or less.
      DEFAULT: DSMAX=0.1 times the cylinder thickness
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("DSMAX", (int, int, float),
                              comment="Maximum step length in body KL,KC")
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, dsmax):
        return super().add(kl, kc, dsmax)

class EABSB(KeywordSequence):
    """
    Local absorption energies EABSB(KPAR,KL,KC) of particles of
    type KPAR in body KL,KC. These values must be larger than
    EABS(KPAR,M), where M is the material of body KL,KC. When
    the particle is moving within body KL,KC, the absorption
    energy EABS(KPAR,M) is temporarily set equal to EABSB(KPAR,
    KL,KC). Thus, the simulation of the particle history is
    discontinued when the energy becomes less than EABSB(KPAR,
    KL,KC). This feature can be used, e.g., to reduce the
    simulation work in regions of lesser interest.
      DEFAULTS: EABSB(KPAR,KL,KC)=EABS(KPAR,M)  (no action)
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("EABSB", (int, int, float, float, float),
                              comment="Local EABSB(1:3) in body KL,KC")
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, eabs1, eabs2, eabs3):
        return super().add(kl, kc, eabs1, eabs2, eabs3)

class IFORCE(KeywordSequence):
    """
    Activates forcing of interactions of type ICOL of particles
    KPAR in cylinder KC of layer KL. FORCER is the forcing
    factor, which must be larger than unity. WLOW and WHIG are
    the lower and upper limits of the weight window where
    interaction forcing is applied. When several interaction
    mechanisms are forced in the same body, the effective weight
    window is set equal to the intersection of the windows for
    these mechanisms.
      DEFAULT: no interaction forcing
    
    If the mean free path for real interactions of type ICOL is
    MFP, the program will simulate interactions of this type
    (real or forced) with an effective mean free path equal to
    MFP/FORCER.
    
    TRICK: a negative input value of FORCER, -FN, is assumed to
    mean that a particle with energy E=EPMAX should interact,
    on average, +FN times in the course of its slowing down to
    rest, for electrons and positrons, or along a mean free
    path, for photons. This is very useful, e.g., to generate
    x-ray spectra from bulk samples.
    """

    def __init__(self, maxlength=120000):
        keyword = TypeKeyword("IFORCE", (int, int, int, int, float, float, float),
                              comment="KL,KC,KPAR,ICOL,FORCER,WLOW,WHIG")
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, kpar, icol, forcer, wlow, whig):
        return super().add(kl, kc, kpar, icol, forcer, wlow, whig)

class IBRSPL(KeywordSequence):
    """
    Activates bremsstrahlung splitting in cylinder KC of layer
    KL for electrons and positrons with weights in the window
    (WLOW,WHIG) where interaction forcing is applied. The
    integer IBRSPL is the splitting factor.
      DEFAULT: no bremsstrahlung splitting
    
    Note that bremsstrahlung splitting is applied in combination
    with interaction forcing and, consequently, it is activated
    only in those bodies where interaction forcing is active.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("IBRSPL", (int, int, float),
                              comment="KL,KC,splitting factor")
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, ibrspl):
        return super().add(kl, kc, ibrspl)

class IXRSPL(KeywordSequence):
    """
    Splitting of characteristic x rays emitted in cylinder KC
    of layer KL, from any element. Each unsplit x ray with
    ILB(2)=2 (i.e., of the second generation) when extracted
    from the secondary stack is split into IXRSPL quanta. The
    new, lighter, quanta are assigned random directions
    distributed isotropically.
      DEFAULT: no x-ray splitting
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("IXRSPL", (int, int, float),
                              comment="KL,KC,splitting factor")
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, ixrspl):
        return super().add(kl, kc, ixrspl)

class IWOODC(TypeKeyword):
    """
    Photons are transported freely across the system using an
    inverse mean free path, SUP, that is larger than the actual
    inverse mean free path, IMFP. The event at the end of each
    free flight may be either a real interaction (ICOL=1 to 4)
    or a delta interaction (ICOL=7). The latter occur with
    probability 1 - IMFP/SUP, so that the simulation remains
    unbiased.
    The method is very effective for the calculation of dose,
    preferably combined with interaction forcing.
      DEFAULT: off
    """

    def __init__(self):
        super().__init__('IWOODC', (bool,),
                         comment='Delta scattering is turned on')

    def set(self, on):
        super().set(on)

    def write(self, fileobj):
        value = self.get()[0]
        if not value:
            return

        line = self._create_line(self.name, [], self.comment)
        fileobj.write(line + os.linesep)

class NBZ(TypeKeyword):
    """
    Number of bins for the Z-coordinate; .LE. 200.
        DEFAULT: NBZ=100
    """

    def __init__(self):
        super().__init__('NBZ', (int,),
                         comment='No. of bins for the Z-coordinate')

    def set(self, nbz):
        super().set(nbz)

class NBR(TypeKeyword):
    """
    Number of bins for the radial variable, R=SQRT(X*X+Y*Y); .LE. 200.
        DEFAULT: NBR=100
    """

    def __init__(self):
        super().__init__('NBR', (int,), comment='No. of radial bins')

    def set(self, nbr):
        super().set(nbr)

class NBTL(TypeKeyword):
    """
    Limits of the interval where track-length distributions of
    primary particles are tallied. Number of track-length bins;
    .LE. 200.
      DEFAULT: TLMIN=0, TLMAX=5*RANGE(EPMAX,KPARP,1), NBTL=200
    """

    def __init__(self):
        super().__init__('NBTL', (float, float, int),
                         comment='Track-length interval and no. of TL-bins')

    def set(self, tlmin, tlmax, nbtl):
        super().set(tlmin, tlmax, nbtl)

class EMERGP(TypeKeyword):
    """
    The program can generate the PDFs of the coordinates (X,Y) of points
    where trajectories of emerging particles intersect the upper and
    lower planes of the geometry definition (upbound and downbound
    particles, respectively), and the PDF of the radial distance of the
    intersections, R=SQRT(X*X+Y*Y).
    
    Activates the generation of the distributions of trajectory
    crossings with the upper and lower planes. RADM is the
    radius of the scoring region, and NBRE is the number of
    bins of the radial distributions. The (X,Y) distributions
    are tallied on a square of side 2*RADM, with (2*NBRE)**2
    bins.
      DEFAULT: None
    """

    def __init__(self):
        super().__init__('EMERGP', (float, int),
                         comment='Radius of the scoring region, no. of bins')

    def set(self, radm, nbre):
        super().set(radm, nbre)

class EnergyDepositionDetectorGroup(KeywordGroup):
    """
    Each energy-deposition detector consists of a set of active bodies,
    which must have been defined as parts of the geometry. The output
    spectrum is the distribution of absorbed energy (per primary shower)
    in the active bodies.
    
             *** WARNING: The energy-deposition spectrum may be strongly
             biased when interaction forcing is applied, even outside the
             detector bodies.
    """

    def __init__(self):
        self.ENDETC = penelope_keywords.ENDETC()
        self.EDSPC = penelope_keywords.EDSPC()

        keyword = TypeKeyword('EDBODY', (int, int), comment='Active cylinder')
        self.EDBODY = KeywordSequence(keyword, maxlength=5000)

    def get_keywords(self):
        return (self.ENDETC, self.EDSPC, self.EDBODY)

    def set(self, el, eu, nbe, spectrum_filename=None, cylinders=None):
        self.ENDETC.set(el, eu, nbe)
        self.EDSPC.set(spectrum_filename)
        self._set_keyword_sequence(self.EDBODY, cylinders)

class EnergyDepositionDetectors(KeywordSequence):

    def __init__(self, maxlength=25):
        keyword = EnergyDepositionDetectorGroup()
        super().__init__(keyword, maxlength)

    def add(self, el, eu, nbe, spectrum_filename=None, cylinders=None):
        return super().add(el, eu, nbe, spectrum_filename, cylinders)

class DOSE2D(KeywordSequence):
    """
    The program will tally 2D, depth-radius, dose and deposited
    charge distributions in the body KL,KC (i.e. the cylinder
    KC of layer KL). The numbers NZ and NR of Z- and R-bins have
    to be specified by the user, they must be in the interval
    from 1 to 200. Up to ten different bodies can be selected,
    a separate line for each body.
      DEFAULT: off
    """

    def __init__(self, maxlength=10):
        keyword = TypeKeyword('DOSE2D', (int, int, int, int),
                              comment='Tally distributions in KL,KC with NZ,NR bins')
        super().__init__(keyword, maxlength)

    def add(self, kl, kc, nz, nr):
        super().add(kl, kc, nz, nr)





"""
Separators used across different PENELOPE main programs.
"""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.separator import Separator

# Globals and constants variables.

DOT = Separator('.')
"""Section separator."""

SOURCE_DEFINITION = Separator('>>>>>>>> Source definition.')
"""Section for source definition."""

MATERIAL = Separator('>>>>>>>> Material data and simulation parameters.')
"""Section for material(s)."""

INTERACTION_FORCING = Separator('>>>>>>>> Interaction forcing.')
"""Section for interaction forcing(s)."""

BREMSSTRAHLUNG_SPLITTING = Separator('>>>>>>>> Bremsstrahlung splitting.')
"""Section for Bremsstralung splitting."""

XRAY_SPLITTING = Separator('>>>>>>>> X-ray splitting.')
"""Section for characteristic x ray splitting."""

EMERGING_PARTICLES = Separator('>>>>>>>> Emerging particles. Energy and angular distributions.')
"""Section for emerging particles."""

ENERGY_DEPOSITON_DETECTORS = Separator('>>>>>>>> Energy-deposition detectors (up to 25).')
"""Section for energy deposition detectors."""

JOB_PROPERTIES = Separator('>>>>>>>> Job properties.')
"""Section for job properties."""

END = Separator('Ends the reading of input data', 'END')
"""End separator"""



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging

# Third party modules.
import pyxray

# Local modules.
from pypenelopetools.penepma.utils import convert_xrayline_to_izs1s200

# Globals and constants variables.

class Testutils(unittest.TestCase):

    def testconvert_xrayline_to_izs1s200(self):
        xrayline = pyxray.xray_line(29, 'Ka2')
        izs1s200 = convert_xrayline_to_izs1s200(xrayline)
        self.assertEqual(29010300, izs1s200)

        xrayline = pyxray.xray_line(29, 'Ka')
        self.assertRaises(ValueError, convert_xrayline_to_izs1s200, xrayline)

if __name__ == '__main__': #pragma: no cover
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



"""
Material definition.

Example:
    Create a material definition for copper::

        mat = Material('copper', {29: 1.0}, 8.9)
        with open('copper.mat.in', 'w') as fp:
            mat.write_input(fp)
    
    Using the ``material`` program of PENELOPE in the ``pendbase`` folder, run::
    
        material.exe < copper.mat.in
    
    There should now be a ``copper.mat`` in the ``pendbase`` folder.
"""

# Standard library modules.
import os

# Third party modules.
import pyxray

# Local modules.

# Globals and constants variables.
FILENAME_MAXLENGTH = 20

class Material(object):
    """
    Creates a new material.
    
    Args:
        name (str): Name of material.
        composition (dict): Composition in weight fraction.
            The composition is specified by a dictionary.
            The keys are atomic numbers and the values are weight fraction 
            between ]0.0, 1.0].
        density_g_per_cm3 (float): Material's density in g/cm3.
        mean_excitation_energy_eV (float, optional): mean excitation energy.
            If ``None``, it will be calculated by PENELOPE.
        oscillator_strength_fcb (float, optional): 
            oscillator strength of plasmon excitation.
            If ``None``, it will be estimated by PENELOPE.
        plasmon_energy_wcb_eV (float, optional): energy of plasmon excitation in eV.
            If ``None``, it will be estimated by PENELOPE.
            
    Attributes:
        name (str): Name of material.
        composition (dict): Composition in weight fraction.
            The composition is specified by a dictionary.
            The keys are atomic numbers and the values are weight fraction 
            between ]0.0, 1.0].
        density_g_per_cm3 (float): Material's density in g/cm3.
        mean_excitation_energy_eV (float, optional): mean excitation energy.
            If ``None``, it will be calculated by PENELOPE.
        oscillator_strength_fcb (float, optional): 
            oscillator strength of plasmon excitation.
            If ``None``, it will be estimated by PENELOPE.
        plasmon_energy_wcb_eV (float, optional): energy of plasmon excitation in eV.
            If ``None``, it will be estimated by PENELOPE.
    """

    def __init__(self, name, composition, density_g_per_cm3,
                 mean_excitation_energy_eV=None,
                 oscillator_strength_fcb=None,
                 plasmon_energy_wcb_eV=None):
        self.name = name
        self.filename = name[:16] + '.mat'
        self.composition = composition.copy()
        self.density_g_per_cm3 = float(density_g_per_cm3)
        self.mean_excitation_energy_eV = mean_excitation_energy_eV
        self.oscillator_strength_fcb = oscillator_strength_fcb
        self.plasmon_energy_wcb_eV = plasmon_energy_wcb_eV

    def __repr__(self):
        return '<{0}({1})>'.format(self.__class__.__name__, self.name)

    @classmethod
    def read_input(cls, fileobj):
        """
        Reads the input file created by this class 
        (see :meth:`write_input <.Material.write_input>`).
        
        Args:
            fileobj (file object): file object opened with read access.
            
        Returns:
            Material: new material.
        """
        composition_option = fileobj.readline().strip()
        assert composition_option == "1"

        name = fileobj.readline().strip()

        element_count = int(fileobj.readline())
        if element_count == 1:
            z = int(fileobj.readline())
            composition = {z: 1.0}
        else:
            wf_option = fileobj.readline().strip()
            assert wf_option == "2"

            composition = {}
            for _ in range(element_count):
                z, wf = fileobj.readline().split()
                composition[int(z)] = float(wf)

        mean_excitation_option = fileobj.readline().strip()
        if mean_excitation_option == "1":
            mean_excitation_energy_eV = float(fileobj.readline())
        else:
            mean_excitation_energy_eV = None

        density_g_per_cm3 = float(fileobj.readline())

        oscillator_option = fileobj.readline().strip()
        if oscillator_option == "1":
            oscillator_strength_fcb, plasmon_energy_wcb_eV = map(float, fileobj.readline().split())
        else:
            oscillator_strength_fcb = plasmon_energy_wcb_eV = None

        filename = fileobj.readline()

        material = cls(name, composition, density_g_per_cm3,
                       mean_excitation_energy_eV,
                       oscillator_strength_fcb,
                       plasmon_energy_wcb_eV)
        material.filename = filename

        return material

    @classmethod
    def read_material(cls, fileobj):
        """
        Reads a PENELOPE generated material file (.mat).
        
        Args:
            fileobj (file object): file object opened with read access.
            
        Returns:
            Material: new material.
        """
        first_line = fileobj.readline().strip()
        assert first_line[:8] == 'PENELOPE'

        line = fileobj.readline()
        _label, value = line.split()
        name = value.strip()

        line = fileobj.readline()
        _label, value = line.split('=')
        value, _unit = value.split()
        density_g_per_cm3 = float(value)

        line = fileobj.readline()
        _label, value = line.split('=')
        element_count = int(value)

        composition = {}
        totalatomicmass = 0.0
        for _ in range(element_count):
            line = fileobj.readline()
            part_z, part_af = line.split(',')

            _label, value = part_z.split('=')
            z = int(value)

            _label, value = part_af.split('=')
            atomicfraction = float(value)
            atomicmass = atomicfraction * pyxray.element_atomic_weight(z)

            composition[z] = atomicmass
            totalatomicmass += atomicmass

        for z, atomicmass in composition.items():
            composition[z] = atomicmass / totalatomicmass

        line = fileobj.readline()
        _label, value = line.split('=')
        value, _unit = value.split()
        mean_excitation_energy_eV = float(value)

        line = fileobj.readline()
        label, _value = line.split('=')
        assert label.strip() == 'Number of oscillators'

        line = fileobj.readline()
        _, value1, _, value2, _, _ = line.split()
        oscillator_strength_fcb = float(value1)
        plasmon_energy_wcb_eV = float(value2)

        return cls(name, composition, density_g_per_cm3,
                   mean_excitation_energy_eV,
                   oscillator_strength_fcb,
                   plasmon_energy_wcb_eV)

    def _create_lines(self):
        """
        Creates the lines of the input file of the material program.
        """
        lines = []

        elements_count = len(self.composition)
        density = self.density_g_per_cm3

        # Select the manual composition option.
        lines.append("1")

        # Material name.
        lines.append(str(self.name))

        lines.append(str(elements_count))

        # Case more than one element.
        if elements_count > 1:
            lines.append("2") # Select weight fraction option.

            for z, wf in self.composition.items():
                lines.append('{0:d} {1:.4f}'.format(z, wf))

        elif elements_count == 1:
            z = list(self.composition.keys())[0]
            lines.append('{0:d}'.format(z))

        else:
            raise ValueError("No elements are defined in the material.")

        # Mean excitation energy.
        if self.mean_excitation_energy_eV is not None:
            lines.append("1")
            lines.append("{0:f}".format(self.mean_excitation_energy_eV))
        else:
            lines.append("2")

        # Density
        lines.append("{0:f}".format(density))

        # Fcb and Wcb values.
        if self.oscillator_strength_fcb is not None and \
                self.plasmon_energy_wcb_eV is not None:
            lines.append("1")
            lines.append("{0:f} {1:f}".format(self.oscillator_strength_fcb,
                                              self.plasmon_energy_wcb_eV))
        else:
            lines.append("2")

        # set the name and path of the output file.
        lines.append(self.filename)

        return lines

    def write_input(self, fileobj):
        """
        Writes the input file to create this material.
        The material program should be called with this input file as standard input::
        
            material.exe < material.in
        
        Args:
            fileobj (file object): file object opened with write access.
        """
        lines = self._create_lines()
        fileobj.write(os.linesep.join(lines))

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, filename):
        if len(filename) > FILENAME_MAXLENGTH:
            raise ValueError("Filename is too long. Maximum {0} characters"
                             .format(FILENAME_MAXLENGTH))
        self._filename = filename

VACUUM = Material('Vacuum', {}, 0.0)
"""
Material representing vacuum.
"""